Package: IsoCorrectoR
Title: Correction for natural isotope abundance and tracer purity in MS
        and MS/MS data from stable isotope labeling experiments
Version: 1.29.0
Authors@R: c(person("Christian","Kohler",email="christian.kohler@ur.de",role=c("cre","aut")),
  person("Paul","Heinrich",email="paul.heinrich@ur.de",role=c("aut")))
Imports: dplyr, magrittr, methods, quadprog, readr, readxl, stringr,
        tibble, tools, utils, pracma, WriteXLS
Description: IsoCorrectoR performs the correction of mass spectrometry data from stable isotope labeling/tracing metabolomics experiments with regard to natural isotope abundance and tracer impurity. Data from both MS and MS/MS measurements can be corrected (with any tracer isotope: 13C, 15N, 18O...), as well as ultra-high resolution MS data from multiple-tracer experiments (e.g. 13C and 15N used simultaneously). See the Bioconductor package IsoCorrectoRGUI for a graphical user interface to IsoCorrectoR. NOTE: With R version 4.0.0, writing correction results to Excel files may currently not work on Windows. However, writing results to csv works as before.
Depends: R (>= 3.5)
URL: https://genomics.ur.de/files/IsoCorrectoR/
License: GPL-3
LazyData: TRUE
NeedsCompilation: no
biocViews: Software, Metabolomics, MassSpectrometry, Preprocessing,
        ImmunoOncology
RoxygenNote: 6.1.1
Suggests: IsoCorrectoRGUI, knitr, rmarkdown, testthat, BiocStyle
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/IsoCorrectoR
git_branch: devel
git_last_commit: 9553234
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-10-30
Packaged: 2025-10-30 21:39:19 UTC; biocbuild
Author: Christian Kohler [cre, aut],
  Paul Heinrich [aut]
Maintainer: Christian Kohler <christian.kohler@ur.de>

# Generated by roxygen2: do not edit by hand

export(IsoCorrection)
import(quadprog)
import(utils)
importFrom(WriteXLS,WriteXLS)
importFrom(dplyr,arrange_)
importFrom(dplyr,last)
importFrom(magrittr,'%>%')
importFrom(methods,is)
importFrom(pracma,lsqlincon)
importFrom(readr,read_csv)
importFrom(readxl,read_excel)
importFrom(stringr,str_c)
importFrom(stringr,str_detect)
importFrom(stringr,str_extract)
importFrom(stringr,str_extract_all)
importFrom(stringr,str_replace)
importFrom(stringr,str_split)
importFrom(tibble,as_tibble)
importFrom(tools,file_ext)
importFrom(utils,packageDescription)
importFrom(utils,packageVersion)
importFrom(utils,write.csv)

# Calculation of the probability matrix for Ultra High Resolution (UHR) 
# data correction

ProbUltraHighRes <- function(MoleculeInfo,MoleculesTotal,ElementInfo,CorrectTracerElementCore,CorrectTracerImpurity,
                             CalculationThreshold_UHR, verbose) {
  
  if(verbose){message(date(), " :: ProbUltraHighRes")
  message(date(), " :: CorrectTracerElementCore: ", CorrectTracerElementCore)
  message(date(), " :: CorrectTracerImpurity: ", CorrectTracerImpurity)
  message(date(), " :: CalculationThreshold_UHR: ", CalculationThreshold_UHR)}
  
  CombinedProbList <- list()
  
  for (MoleculeNo in seq_len(MoleculesTotal)) {
    
    MoleculeData <- MoleculeInfo[[MoleculeNo]]
    
    NatAbuImpurityList <- IsoCombinationsUHR(MoleculeData=MoleculeData,
                                                 ElementInfo=ElementInfo,
                                                 CorrectTracerElementCore=CorrectTracerElementCore,
                                                 CorrectTracerImpurity=CorrectTracerImpurity,
                                                 verbose=verbose)
    
    CumProbList <- TracerCountProbabilitiesUHR(MoleculeData=MoleculeData,
                                               NatAbuImpurityList=NatAbuImpurityList,
                                               verbose=verbose)
    
    CombinedProb <- TracersCombinedProbabilityUHR(MoleculeData=MoleculeData,
                                                  CumProbList=CumProbList,
                                                  CalculationThreshold_UHR=CalculationThreshold_UHR)
    
    CombinedProbList[[MoleculeNo]] <- CombinedProb
    
  }

  names(CombinedProbList) <- names(MoleculeInfo)
  
  if(verbose){message(date(), " :: leaving ProbUltraHighRes()")}
  
  return(CombinedProbList)
  
}

# COMPUTE TRACER IMPURITY COMBINATIONS

# Tracer purity is the probability that a tracer element atom in the 
# tracer substrate that should be labelled actually is labelled. 
# E.g.  a 99.0% isotopic purity 1,2-13C-Glucose has a 99.0% chance for 
# each of its carbon atom positions 1 and 2 of containing a 13C. 
# Thus, there is a 1.0% chance for each of those carbons that they are not 13C.
# Consequently, molecules that contain the tracer
# isotope due to metabolic transfer from the tracer substrate and not due to 
# natural abundance have a certain chance of contributing to labelling states
# with less tracer incorporated. This is due to the decrease in mass shift 
# associated with tracer impurity (e.g. 12C instead of 13C at a carbon position).
# In the following, for each labelling state the probability of having a 
# certain number of 'impure' tracer positions as well as the associated 
# (negative) mass shift are is calculated. 
# The maximum amount of impure tracer positions possible depends on the number
# of label present in a given transition.

IsoCombinationsTracerImpurity <- function(MoleculeFragmentData, ElementInfo, Transitions, NumberTransitions,
                                          IDTracer, LabelPresent, CalculationThreshold) { 

  ProbTracerImpurityList <- list()
  MassShiftTracerImpurityList <- list()
  NumberImpurityCombEff <- list()
  
  for (TransitionNo in seq_len(NumberTransitions)) {
    
    tmpProbTracerImpurity_Array <- vector()
    tmpMassShiftTracerImpurity_Array <- vector()
    tmpNumberImpurityCombEff <- vector()
    
    for (ImpureTracer in 0:LabelPresent[TransitionNo]) {
      
      ProbImpurity <- choose(LabelPresent[TransitionNo], ImpureTracer) * (ElementInfo[[IDTracer]][["Tracer.purity"]]^(LabelPresent[TransitionNo] - 
                                                                                                                        ImpureTracer)) * ((1 - ElementInfo[[IDTracer]][["Tracer.purity"]])^ImpureTracer)
      
      if (ProbImpurity > CalculationThreshold) {
        tmpProbTracerImpurity_Array[ImpureTracer + 1] <- ProbImpurity
        tmpMassShiftTracerImpurity_Array[ImpureTracer + 1] <- (-1) * ImpureTracer * ElementInfo[[IDTracer]][["Tracer.isotope.mass.shift"]]
        NumberImpurityCombEff[TransitionNo] <- ImpureTracer
      } else {
        ImpureTracer <- LabelPresent + 1
      }  #ProbImpurity > CalculationThreshold
    }  #ImpureTracer
    
    ProbTracerImpurityList[[TransitionNo]] <- tmpProbTracerImpurity_Array
    MassShiftTracerImpurityList[[TransitionNo]] <- tmpMassShiftTracerImpurity_Array
  }  #TransitionNo
  
  names(ProbTracerImpurityList) <- rownames(Transitions)
  names(MassShiftTracerImpurityList) <- rownames(Transitions)
  names(NumberImpurityCombEff) <- rownames(Transitions)
  
  return(list("ProbTracerImpurityList"=ProbTracerImpurityList,
              "MassShiftTracerImpurityList"=MassShiftTracerImpurityList,
              "NumberImpurityCombEff"=NumberImpurityCombEff))
 
}

# COMPUTE ISOTOPE COMBINATIONS OF TRACER ELEMENT

IsoCombinationsTracer <- function(MoleculeFragmentData, Fragment, ElementInfo, Transitions, NumberTracers, CorrectTracerImpurity, CorrectTracerElementCore, CalculationThreshold) {
  ProbTracerList <- list()
  MassShiftTracerList <- list()
  NumberIsoCombTracerEff <- list()
  IsotopesTracer <- list()

  NumberTransitions <- nrow(Transitions)
  NegIsoTracer <- MoleculeFragmentData[["NegIsoTracer"]]
  nTracerMax <- MoleculeFragmentData[["nTracerMax"]]
  MaxLabel <- MoleculeFragmentData[["MaxLabel"]]

  LabelPresent <- rep(0, nrow(Transitions))

  if (NumberTracers > 0) {
    IDTracer <- MoleculeFragmentData[["IDTracer"]]
    Element <- IDTracer
    LabelPresent <- vector()
    nTracer <- vector()

    for (TransitionNo in seq_len(NumberTransitions)) {

      # x is passed to the daughter function 'IsoCombinations'

      x <- TransitionNo

      # ElementInfo[[Element]][[2]] is the tracer isotope mass shift.
      # Dividing by this transforms the mass shift of a transition into
      # the number of label present.

      LabelPresent[x] <- Transitions[x, Fragment] / ElementInfo[[Element]][[2]]

      nTracer[x] <- as.numeric(nTracerMax - LabelPresent[x])

      # Do not constrain PlacesToAssign if tracer purity correction is active,
      # as tracer purity produces negative mass shifts which leads to similar
      # effects as NegIsoTracer.

      if (CorrectTracerImpurity || NegIsoTracer == 1) {
        PlacesToAssign <- as.numeric(ElementInfo[[Element]][[2]]) * MaxLabel
      } else {
        PlacesToAssign <- as.numeric(ElementInfo[[Element]][[2]] * (MaxLabel - LabelPresent[x]))
      }

      # Correct molecule core or not

      if (CorrectTracerElementCore) {
        AvailablePlacesTotal <- nTracer[x]
      } else if (!CorrectTracerElementCore) {
        AvailablePlacesTotal <- as.numeric(nTracerMax) - as.numeric(MaxLabel)
      } else {
        stop(date(), " :: Invalid value for `CorrectTracerElementCore`: ", CorrectTracerElementCore)
      }

      # Definition of the IsoCluster variable that is transferred to the
      # daughter function IsoCombinations() that computes the actual
      # isotope combinations

      if (PlacesToAssign <= AvailablePlacesTotal) {
        IsoCluster <- PlacesToAssign
      } else {
        IsoCluster <- AvailablePlacesTotal
      }

      IsoCombinationsResultList <- IsoCombinations(Fragment = Fragment, x = x, Element = Element, ElementInfo = ElementInfo, AvailablePlacesTotal = AvailablePlacesTotal, IsoCluster = IsoCluster, CalculationThreshold = CalculationThreshold)
      NumberIsoComb <- IsoCombinationsResultList[["NumberIsoComb"]]
      tmpProbTracer_Array <- vector()
      tmpMassShiftTracer_Array <- vector()
      tmpIsotopesTracer <- list()
      for (i in seq_len(NumberIsoComb)) {
        tmpProbTracer_Array[i] <- IsoCombinationsResultList[["ProbArray"]][[i]]
        tmpMassShiftTracer_Array[i] <- IsoCombinationsResultList[["MassShiftArray"]][[i]]
      } # i

      ProbTracerList[[x]] <- tmpProbTracer_Array
      MassShiftTracerList[[x]] <- tmpMassShiftTracer_Array
      NumberIsoCombTracerEff[[x]] <- NumberIsoComb
      IsotopesTracer[[x]] <- IsoCombinationsResultList[["Isotopes"]]
    } # TransitionsNo (represented by x)

    names(ProbTracerList) <- rownames(Transitions)
    names(MassShiftTracerList) <- rownames(Transitions)
    names(NumberIsoCombTracerEff) <- rownames(Transitions)
    names(IsotopesTracer) <- rownames(Transitions)
  } else {
    LabelPresent <- rep(0, NumberTransitions)
  }

  return(list(
    "ProbTracerList" = ProbTracerList,
    "MassShiftTracerList" = MassShiftTracerList,
    "NumberIsoCombTracerEff" = NumberIsoCombTracerEff,
    "IsotopesTracer" = IsotopesTracer,
    "LabelPresent" = LabelPresent
  ))
}

# Calculate combined tracer probabilites for UHR

# In a way similar to the function FragmentsCombinedProbability in the normal 
# resolution mode, this section gives the probability that a given labeling state
# y produces the same isotope pattern as a labeling state x by natural abundance 
# (and tracer impurity). To this end, the probability that the number of tracer isotopes in 
# transition x is produced by transition y for each tracer element is 
# multiplied with the other tracer element probabilities. 
# The variable CalculationThreshold_UHR constraints
# the transitions y for which the contribution to x is calculated to 
# CalculationThreshold_UHR which is +/- the total amount of 
# additional/substracted label
# (all tracers summed up) compared to x.

TracersCombinedProbabilityUHR <- function(MoleculeData,CumProbList,CalculationThreshold_UHR) {

  NumberTracers <- length(MoleculeData[[1]][["Tracer"]])
  
  Transitions <- MoleculeData[["Transitions"]]
  NumberTransitions <- nrow(Transitions)
  
  CombinedProb <- matrix(ncol = NumberTransitions, nrow = NumberTransitions)
  
  for (x in seq_len(NumberTransitions)) {
    if (CalculationThreshold_UHR > 0) {
      MultipleTracerProbVector <- vector()
      for (y in seq_len(NumberTransitions)) {
        if (((Transitions[x, NumberTracers + 1] - Transitions[y, NumberTracers + 1])^2)^0.5 <= CalculationThreshold_UHR) {
          Prob <- 1
          for (TracerNo in seq_len(NumberTracers)) {
            Prob <- Prob * CumProbList[[TracerNo]][[Transitions[y, TracerNo] + 1]][Transitions[x, TracerNo] + 1]
          }  #TracerNo
        } else {
          Prob <- 0
        }
        MultipleTracerProbVector[y] <- Prob
      }  #y
      CombinedProb[x, ] <- MultipleTracerProbVector
      rm(MultipleTracerProbVector)
    } else {
      for (y in seq_len(NumberTransitions)) {
        Prob <- 1
        for (TracerNo in seq_len(NumberTracers)) {
          Prob <- Prob * CumProbList[[TracerNo]][[Transitions[y, TracerNo] + 1]][Transitions[x, TracerNo] + 1]
        }  #TracerNo
        MultipleTracerProbVector[y] <- Prob
      }  #y
      CombinedProb[x, ] <- MultipleTracerProbVector
      rm(MultipleTracerProbVector)
    }  #if(CalculationThreshold_UHR>0)
  }  #x
  colnames(CombinedProb) <- rownames(Transitions)
  rownames(CombinedProb) <- rownames(Transitions)
 
  return(CombinedProb)
   
}

# Calculation of probability that a given labelling state produces the mass shift of another labelling state

# In the function 'FragmentsCombinedProbability' the final
# probability matrix CombFragmentProb is produced. In this matrix, for each
# transition y of a molecule the probability that it produces a transition x
# of higher mass through natural isotope abundance is given.
# This is done by matching the mass shifts from
# FragmentMassShift and CumProbList, as described before for the
# function 'MassShiftTransitions'. For MS/MS data, the
# probability that the product ion y produces the product ion
# mass shift of transition x and and the probability that neutral
# loss y produces the neutral loss mass shift of transition x
# are multiplied. This way the overall probability that
# transition y produces transition x through natural isotope abundance
# is yielded.

#' @importFrom stringr str_detect
#' @importFrom magrittr '%>%'
#' 
FragmentsCombinedProbability <- function(MoleculeInfo, MoleculeNo, CumProbList, FragmentMassShift) {
  
  NumberFragments <- stringr::str_detect(names(MoleculeInfo[[MoleculeNo]]), "Fragment") %>% sum
  Transitions <- MoleculeInfo[[MoleculeNo]][["Transitions"]]
  NumberTransitions <- nrow(Transitions)
  
  CombinedProb <- matrix(NA, nrow = NumberTransitions, ncol = NumberTransitions)
  colnames(CombinedProb) <- rownames(Transitions)
  rownames(CombinedProb) <- rownames(Transitions)
  
  if (NumberFragments == 2) {
    for (x in seq_len(NumberTransitions)) {
      for (y in seq_len(NumberTransitions)) {
        CombinedProb[x, y] <- as.data.frame(CumProbList[[1]])[y, FragmentMassShift[[1]][x] + 1] * as.data.frame(CumProbList[[2]])[y, FragmentMassShift[[2]][x] + 1]
      }  #y
    }  #x
  } else if (NumberFragments == 1) {
    for (x in seq_len(NumberTransitions)) {
      for (y in seq_len(NumberTransitions)) {
        CombinedProb[x, y] <- as.data.frame(CumProbList[[1]])[y, FragmentMassShift[[1]][x] + 1]
      }  #y
    }  #x
  }
  
  return(CombinedProb)
  
}

# Element Parameter Extraction

#' @importFrom stringr str_split
#' @importFrom tibble as_tibble
#' @importFrom magrittr '%>%'
#' 
ElementInfoExtraction <- function(ElementData, UltraHighRes, logEnvironment, verbose) {
  if(verbose){message(date(), " :: processing element file ...")}
  
  # Using two nested loops, the isotope information of each element in column 2 is extracted by splitting it using the delimiters '/'
  # (isotopes1/isotope2...) and '_' (abundance-mass shift of given isotope).
  
  ElementList <- list()
  
  for (Element in seq_len(nrow(ElementData))) {
    
    Isotopes <- stringr::str_split(ElementData[Element, 2], pattern = "/", simplify = TRUE)
    
    ElementArray <- matrix(NA, nrow = length(Isotopes), ncol = 2)
    colnames(ElementArray) <- c("IsotopeAbundance", "MassShift")
    for (Isotope in seq_len(length(Isotopes))) {
      
      IsoAbundanceMassShift <- stringr::str_split(Isotopes[Isotope], pattern = "_", simplify = TRUE)
      
      for (i in seq_len(ncol(ElementArray))) {
        ElementArray[Isotope, i] <- as.numeric(IsoAbundanceMassShift[i])
      }
      
    }  #Isotope
    
    # Sort Isotopes of each element in ElementArray ascending by probability. This is required to be able to later calculate probabilities in descending
    # order, making it possible to stop the calculation at a defined threshold without losing higher probability values.
    
    ElementArray_df <- as.data.frame(ElementArray)
    
    ElementArray_df <- ElementArray_df[order(ElementArray_df[["IsotopeAbundance"]], decreasing = TRUE),] 
    
    ElementArray_tbl <- tibble::as_tibble(ElementArray_df)
    
    tmp.list <- list()
    tmp2.list <- list()
    tmp.list[[1]] <- ElementArray_tbl
    tmp2.list[[1]] <- tmp.list
    
    # now add rest of ElementData
    for (i in 3:ncol(ElementData)) {
      tmp <- ElementData[Element, i]
      
      # Making NAs of ElementArray 0
      
      if (is.na(tmp)) {
        tmp <- 0
      }
      tmp2.list[[i - 1]] <- tmp
    }  #i
    
    names(tmp2.list) <- c("Isotopes", colnames(ElementData)[3:ncol(ElementData)])
    ElementList[[Element]] <- tmp2.list
    rm(tmp.list)
    rm(tmp2.list)
    
  }  #Element
  
  names(ElementList) <- ElementData[[1]]
  
  checkElementDataLogic(ElementList, UltraHighRes, logEnvironment, verbose=verbose)
  
  if(verbose){message(date(), " :: processing element file [OK]\n")}
  return(ElementList)
}  # end of function

# Analysis of molecule formula with regex

# The strings imported from MoleculeFile containing information on the number 
# of atoms per element are analyzed using regular expressions.  The regular
# expression Characters extracts all characters, the regular expression 
# Numbers extracts all numbers.  TracerElement looks for the tracer element
# identifier.

ParseMoleculeInformation <- function(MoleculeInformation, ElementInfo, UltraHighRes, verbose) {
  
  Characters <- "([A-Za-z]*)"
  Numbers <- "([0-9]+[0-9]|[0-9])"
  TracerElement <- "(Lab)"
  MoleculeFormulaRegex <- "(|Lab)[A-Z][a-z]?[1-9][0-9]*"
    
  if (is.na(MoleculeInformation[1,3])) {
    NumberFragments <- 1
  } else {
    NumberFragments <- 2
  }
  
  # collect features for each individual fragment
  tmpFragmentList <- list()
  
  # FragmentList contains one or two tmpFragmentList()s
  FragmentList <- list()
  
  # The following is done for each Molecule and each fragment 
  # of a given molecule
  for (Fragment in seq_len(NumberFragments)) {
    
    # For each molecule(-fragment), the string containing the element count 
    # information is split into the individual elements using a regex
    
    Elements <- stringr::str_extract_all(MoleculeInformation[1,Fragment + 1], MoleculeFormulaRegex)[[1]]
    
    # Now each element in Elements, consisting of the element ID and the 
    # element count, is again separated into ID and element count using 
    # the regular expressions constructed above. 
    # Additionally, a check is made to identify the 
    # tracer element information ('Lab').
    
    ElementCount.vec <- vector()
    Element.vec <- vector()
    TracerCount.vec <- vector()
    Tracer.vec <- vector()
    
    NumberElementsandTracer <- length(Elements)
    
    for (Element in seq_len(NumberElementsandTracer)) {
      
      # TRUE => current Element is Tracer // FALSE => current Element is no Tracer
      if (Elements[Element] %>% stringr::str_detect(TracerElement)) {
        
        TracerCount.vec <- c(TracerCount.vec, Elements[Element] %>% stringr::str_replace(TracerElement, "") %>% stringr::str_extract(Numbers) %>% 
                               as.numeric)
        Tracer.vec <- c(Tracer.vec, Elements[Element] %>% stringr::str_replace(TracerElement, "") %>% stringr::str_extract(Characters))
      } else {
        
        ElementCount.vec <- c(ElementCount.vec, Elements[Element] %>% stringr::str_extract(Numbers) %>% as.numeric)
        Element.vec <- c(Element.vec, Elements[Element] %>% stringr::str_extract(Characters))
      }  #else
    }  #Element    
    names(ElementCount.vec) <- Element.vec
    names(TracerCount.vec) <- Tracer.vec
    
    tmpFragmentList <- list(Element = ElementCount.vec, Tracer = TracerCount.vec)
    
    NumberElements <- length(ElementCount.vec)
    NumberTracers <- length(TracerCount.vec)
    
    # 'NumberElements' - 'NumberTracers' gives the number of non-tracer elements NumberElementsNonTracer.
    NumberElementsNonTracer <- NumberElements - NumberTracers
    
    if (!UltraHighRes) {
      
      # If a tracer is present in the molecule(-fragment) considered, 
      # the tracer parameters are extracted.  'MaxLabel' is the maximum amount
      # of tracer isotope that is expected to be found in the 
      # molecule(-fragment) due to metabolism while 'nTracerMax' is 
      # the maximum amount of tracer element (labelled or
      # unlabelled) in that same species. 'IDTracer' is the tracer elements ID.
      
      if (NumberTracers > 0) {
        
        MaxLabel <- max(TracerCount.vec)
        IDTracer <- TracerCount.vec %>% which.max %>% names
        
        # This part deals with the possibility of having a tracer element 
        # that shows isotopes with a negative mass shift.  
        # In this case IsoCombinationsMaster()
        # has to calculate `PlacesToAssign` differently for the tracer.
        
        if (sum(ElementInfo[[IDTracer]][["Isotopes"]][[1]][["MassShift"]] < 0) > 0) {
          NegIsoTracer <- 1  # => TRUE
        } else {
          NegIsoTracer <- 2  # => FALSE
        }
        
        names(NegIsoTracer) <- IDTracer
        
        tmp.which <- which(names(ElementCount.vec) %in% names(TracerCount.vec))
        nTracerMax <- as.numeric(ElementCount.vec[tmp.which])
        
        if (length(nTracerMax) == NumberTracers) {
          
          names(nTracerMax) <- names(TracerCount.vec)
          
        }
        
        names(MaxLabel) <- IDTracer
        
        tmpFragmentList[["MaxLabel"]] <- MaxLabel
        tmpFragmentList[["IDTracer"]] <- IDTracer
        tmpFragmentList[["nTracerMax"]] <- nTracerMax
        tmpFragmentList[["NegIsoTracer"]] <- NegIsoTracer
      } else {
        tmpFragmentList[["MaxLabel"]] <- NA
        tmpFragmentList[["IDTracer"]] <- NA
        tmpFragmentList[["nTracerMax"]] <- NA
        tmpFragmentList[["NegIsoTracer"]] <- NA
      }  #if(NumberTracers>0)
      
      # In this section, the non-tracer element parameters are extracted in 
      # the same way as for the tracer element.
      
      ElementsNonTracerList <- list()
      
      if (NumberElementsNonTracer > 0) {
        
        NonTracer <- 1
        
        tmpElementNonTracer.vec <- vector()
        tmpElementNonTracerCount.vec <- vector()
        tmpElementZeroTracer.vec <- vector()
        tmpElementZeroTracerCount.vec <- vector()
        
        for (Element in seq_len(NumberElements)) {
          # tracers exist
          if (NumberTracers > 0) {
            # current element is no tracer
            if (names(ElementCount.vec)[Element] != IDTracer) {
              tmpElementNonTracer.vec <- c(tmpElementNonTracer.vec, names(ElementCount.vec)[Element])
              tmpElementNonTracerCount.vec <- c(tmpElementNonTracerCount.vec, as.numeric(ElementCount.vec[Element]))
            }
          } else {
            tmpElementZeroTracer.vec <- c(tmpElementZeroTracer.vec, names(ElementCount.vec)[Element])
            tmpElementZeroTracerCount.vec <- c(tmpElementZeroTracerCount.vec, as.numeric(ElementCount.vec[[Element]]))
            
          }  #NumberTracers
        }  #Element
        
        names(tmpElementNonTracerCount.vec) <- tmpElementNonTracer.vec
        names(tmpElementZeroTracerCount.vec) <- tmpElementZeroTracer.vec
        
        tmpFragmentList[["NonTracer"]] <- tmpElementNonTracerCount.vec
        tmpFragmentList[["ZeroTracer"]] <- tmpElementZeroTracerCount.vec
        
      } else {
        # NumberElementsNonTracer>0
        
        tmpFragmentList[["NonTracer"]] <- c()
        tmpFragmentList[["ZeroTracer"]] <- c()
        
      }
      
    } else if (UltraHighRes) 
    {
      
      if (NumberTracers > 0) {
        
        # For multiple tracer correction, the gathering of tracer parameters 
        # has to loop through the number of tracers present in a 
        # given molecule (-fragment).
        
        MaxLabel.vec <- vector()
        IDTracer.vec <- vector()
        nTracerMax.vec <- vector()
        
        for (TracerNo in seq_len(NumberTracers)) {
          MaxLabel <- as.numeric(TracerCount.vec[TracerNo])
          IDTracer <- names(TracerCount.vec)[TracerNo]
          
          MaxLabel.vec <- c(MaxLabel.vec, MaxLabel)
          IDTracer.vec <- c(IDTracer.vec, IDTracer)
          
          for (Element in seq_len(NumberElements)) {
            # if current element is a tracer
            if (names(tmpFragmentList[["Element"]][Element]) == IDTracer) 
            {
              nTracerMax <- tmpFragmentList[["Element"]][Element] %>% as.numeric
              nTracerMax.vec <- c(nTracerMax.vec, nTracerMax)
            }  #if
          }  #Element
        }  #TracerNo
        names(MaxLabel.vec) <- IDTracer.vec
        
        if (length(nTracerMax.vec) == NumberTracers) {
          
          names(nTracerMax.vec) <- IDTracer.vec
          
        }
        
        tmpFragmentList[["MaxLabel"]] <- MaxLabel.vec
        tmpFragmentList[["IDTracer"]] <- IDTracer.vec
        tmpFragmentList[["nTracerMax"]] <- nTracerMax.vec
        
        # Gaining natural abundance information associated with the 
        # tracer isotopes for probability calculations
        
        # Number of isotopes per Element
        NumberIso <- unlist(lapply(ElementInfo, function(x) nrow(data.frame(x[["Isotopes"]]))))
        
        ### store information for each individual Tracer Element
        NatAbuTracerList <- list()
        NatAbuBaseList <- list()
        MassShiftTracerList <- list()
        
        MassShiftTracer.vec <- vector()
        
        for (TracerNo in names(TracerCount.vec)) {
          MassShiftTracer <- ElementInfo[[TracerNo]][[2]]
          
          # for each Tracer isotope
          NatAbuTracer.vec <- vector()
          NatAbuBase.vec <- vector()
          
          for (IsotopeNo in seq_len(NumberIso[TracerNo])) {
            
            if (data.frame(ElementInfo[[TracerNo]][[1]])[IsotopeNo, 2] == MassShiftTracer) {
              
              NatAbuTracer.vec <- c(NatAbuTracer.vec, data.frame(ElementInfo[[TracerNo]][[1]])[IsotopeNo, 1])  # 1=>IsotopeAbundance
              
            } else if (data.frame(ElementInfo[[TracerNo]][[1]])[IsotopeNo, 2] == 0) {
              
              NatAbuBase.vec <- c(NatAbuBase.vec, data.frame(ElementInfo[[TracerNo]][[1]])[IsotopeNo, 1])
              
            } else {
              
              if(verbose){message(date(), " :: skipping IsotopeNo ", IsotopeNo, " for Tracer ", TracerNo)}
              
            }
          }  #IsotopeNo
          
          NatAbuTracerList[[TracerNo]] <- NatAbuTracer.vec
          NatAbuBaseList[[TracerNo]] <- NatAbuBase.vec
          MassShiftTracerList[[TracerNo]] <- MassShiftTracer
          
        }  #TracerNo
        
        tmpFragmentList[["NatAbuTracer"]] <- unlist(NatAbuTracerList)
        tmpFragmentList[["NatAbuBase"]] <- unlist(NatAbuBaseList)
        tmpFragmentList[["MassShiftTracer"]] <- unlist(MassShiftTracerList)
      } else {
        if(verbose){message(date(), " :: NO TRACER FOR MOLECULE ", MoleculeInformation[1,1], " AND FRAGMENT #", Fragment)}
      }  #NumberTracers>0
    }  #if(UltraHighRes==1)
    
    FragmentList[[Fragment]] <- tmpFragmentList
  }  # Fragment
  
  names(FragmentList) <- stringr::str_c("Fragment_", rep(seq_len(NumberFragments)))
  
  return(FragmentList)
  
}

# CALCULATION OF THE NON-TRACER ELEMENT COMBINATIONS 

ElementCombinationsNonTracer <- function(ProbElemList, MassShiftElemList, NumberIsoCombEff, 
                                         ElementsNonTracer, NumberElementsNonTracer, CalculationThreshold) {

  # If the number of non-tracer elements is > 1, all isotope combination probabilities of non-tracer
  # elements 1 and 2 of a molecule(-fragment) are multiplied with each other and the corresponding mass shifts are added.  Like in 'IsoCombinations' there
  # is a probability threshold which resets to the outer loop if the probability calculated is below.
  
  CombElemProbList <- list()
  tmpCombElemProb_Array <- vector()
  
  CombElemMassShiftList <- list()
  tmpCombElemMassShift_Array <- vector()
  
  NumberEleComb <- list()
  tmpNumberEleComb <- vector()
  
  if (NumberElementsNonTracer > 1) {
    
    EleComb <- 1
    
    for (IsoCombinationA in seq_len(as.numeric(NumberIsoCombEff[1]))) {
      
      for (IsoCombinationB in seq_len(as.numeric(NumberIsoCombEff[[2]]))) {
        
        tmpCombElemProb <- ProbElemList[[1]][IsoCombinationA] * ProbElemList[[2]][IsoCombinationB]
        
        if (tmpCombElemProb > CalculationThreshold) {
          
          tmpCombElemProb_Array[EleComb] <- tmpCombElemProb
          tmpCombElemMassShift_Array[EleComb] <- MassShiftElemList[[1]][IsoCombinationA] + MassShiftElemList[[2]][IsoCombinationB]
          tmpNumberEleComb <- EleComb
          
          EleComb <- EleComb + 1
        } else {
          IsoCombinationB <- NumberIsoCombEff[[2]] + 1
        }  #tmpCombElemProb > CalculationThreshold
      }  #IsoCombinationB
    }  #IsoCombinationA
    
    CombElemProbList[[2]] <- tmpCombElemProb_Array
    CombElemMassShiftList[[2]] <- tmpCombElemMassShift_Array
    NumberEleComb[[2]] <- tmpNumberEleComb
    
    rm(tmpCombElemProb_Array)
    rm(tmpCombElemMassShift_Array)
    rm(tmpNumberEleComb)
    
  } else if (NumberElementsNonTracer > 0) {
    
    # If the number of non-tracer elements is 1, the probability and mass shift arrays are equal to those generated in 'IsoCombinationsResult'.
    
    tmpCombElemProb_Array[seq_len(NumberIsoCombEff[[1]])] <- ProbElemList[[1]][seq_len(NumberIsoCombEff[[1]])]
    tmpCombElemMassShift_Array[seq_len(NumberIsoCombEff[[1]])] <- MassShiftElemList[[1]][seq_len(NumberIsoCombEff[[1]])]
    tmpNumberEleComb <- NumberIsoCombEff[[1]]
    
    CombElemProbList[[1]] <- tmpCombElemProb_Array
    CombElemMassShiftList[[1]] <- tmpCombElemMassShift_Array
    NumberEleComb[[1]] <- tmpNumberEleComb
    
    rm(tmpCombElemProb_Array)
    rm(tmpCombElemMassShift_Array)
    rm(tmpNumberEleComb)
    
  }  #NumberElementsNonTracer>1
  
  # If there are more than 2 non-tracer elements, the element combinations are calculated in iterations of the following outer loop until all elements have
  # been covered. The algorithm uses the CombElemProbList covering the previously combined element probabilities and multiplies these with the isotope
  # combination probabilities of the next element (ProbElemList).  This is fed back into CombElemProbList at the current element index of the loop). The
  # same is done additively for the mass shift array CombElemMassShiftList.
  
  if (NumberElementsNonTracer > 2) {
    
    tmpCombElemProb_Array <- vector()
    tmpCombElemMassShift_Array <- vector()
    tmpNumberEleComb <- vector()
    
    for (NonTracer in 3:NumberElementsNonTracer) {
      EleComb <- 1
      for (ElementCombination in seq_len(NumberEleComb[[NonTracer - 1]])) {
        for (IsoCombination in seq_len(NumberIsoCombEff[[NonTracer]])) {
          
          tmpCombElemProb <- CombElemProbList[[NonTracer - 1]][[ElementCombination]] * ProbElemList[[NonTracer]][[IsoCombination]]
          if (tmpCombElemProb > CalculationThreshold) {
            
            tmpCombElemProb_Array[EleComb] <- tmpCombElemProb
            tmpCombElemMassShift_Array[EleComb] <- CombElemMassShiftList[[NonTracer - 1]][[ElementCombination]] + MassShiftElemList[[NonTracer]][[IsoCombination]]
            tmpNumberEleComb <- EleComb
            EleComb <- EleComb + 1
            
          } else {
            
            IsoCombination <- NumberIsoCombEff[[NonTracer]] + 1
            
          }  #tmpCombElemProb > CalculationThreshold
        }  #IsoCombination
        
      }  #ElementCombination
      CombElemProbList[[NonTracer]] <- tmpCombElemProb_Array
      CombElemMassShiftList[[NonTracer]] <- tmpCombElemMassShift_Array
      NumberEleComb[[NonTracer]] <- tmpNumberEleComb
      
    }  #NonTracer
  }  #NumberElementsNonTracer>2
  
  return(list("CombElemProbList"=CombElemProbList,
              "CombElemMassShiftList"=CombElemMassShiftList,
              "NumberEleComb"=NumberEleComb))
  
}

# Supporting functions

#' @importFrom methods is
#' @importFrom stringr str_c
#' @importFrom stringr str_detect
#' @importFrom readxl read_excel
#' @importFrom readr read_csv
#' @importFrom tools file_ext

#' 
# Function to check parameters of IsoCorrection() function call

checkIsoCorrectionParameters <- function(MeasurementFile, ElementFile, MoleculeFile, CorrectTracerImpurity, CorrectTracerElementCore, CalculateMeanEnrichment, UltraHighRes, FileOut, FileOutFormat, ReturnResultsObject, CorrectAlsoMonoisotopic, CalculationThreshold, CalculationThreshold_UHR, logEnvironment, Testmode, verbose) {
  
  # Check filenames
  
  filenames <- list()
  
  filenames[["MoleculeFile"]] <- MoleculeFile
  filenames[["ElementFile"]] <- ElementFile
  filenames[["MeasurementFile"]] <- MeasurementFile
  filenames[["FileOut"]] <- FileOut
  
  for (name in names(filenames)) {
    
    if (is.na(filenames[[name]]) || filenames[[name]] == "") {
      
      notification <- paste0("No or unsuitable value provided for parameter ", name, ".")
      errorHandler(notification, logEnvironment, "error", verbose=verbose)
      
    }
    
  }
  
  # Check boolean values
  
  booleans <- list()
  
  booleans[["CorrectTracerImpurity"]] <- CorrectTracerImpurity
  booleans[["CorrectTracerElementCore"]] <- CorrectTracerElementCore
  booleans[["CorrectAlsoMonoisotopic"]] <- CorrectAlsoMonoisotopic
  booleans[["UltraHighRes"]] <- UltraHighRes
  booleans[["CalculateMeanEnrichment"]] <- CalculateMeanEnrichment
  booleans[["ReturnResultsObject"]] <- ReturnResultsObject
  booleans[["Testmode"]] <- Testmode
  
  for (name in names(booleans)) {
    
    if (is.na(booleans[[name]]) || (!(booleans[[name]] == TRUE) && !(booleans[[name]] == FALSE) && !(booleans[[name]] == 0) && !(booleans[[name]] == 1))) {
      
      notification <- paste0("No or unsuitable value provided for parameter ", name, ".")
      errorHandler(notification, logEnvironment, "error", verbose=verbose)
      
    }
    
  }
  
  # Check format of output file
  isValid<-FileOutFormat%in%c("xls","csv")
  if(!isValid){
    notification <- paste0("The output format for the result file should be either \"xls\" or \"csv\". Yours was \"", FileOutFormat, "\".")
    errorHandler(notification, logEnvironment, "error", verbose=verbose)
  }
  
  # Check CalculationThreshold values
  
  if (!UltraHighRes) {
    
    if (CalculationThreshold > (10^-2) || CalculationThreshold < 0) {
      
      notification <- paste0("Input parameter 'CalculationThreshold' must be between 0 and 0.01 in normal resolution correction. If unsure, use the default value of 10^-8.")
      errorHandler(notification, logEnvironment, "error", verbose=verbose)
      
    }
    
  } else if (UltraHighRes) {
    
    if (CalculationThreshold_UHR < 0 || (CalculationThreshold_UHR != round(CalculationThreshold_UHR))) {
      
      notification <- paste0("Input parameter 'CalculationThreshold_UHR' must be integer and non-negative in ultra high resolution correction. If unsure, use the default value of 8.")
      errorHandler(notification, logEnvironment, "error", verbose=verbose)
      
    }
    
  }
  
}

# Function to write paths/filenames, correction parameters and errors/warnings to a log file

writeLog <- function(logEnvironment) {
  
  logFile <- file.path(logEnvironment$param$OutputDirectory, logEnvironment$param$LogfileName)
  
  disclaimer <- "IsoCorrectoR is free software and comes without any warranty."
  
  write(paste0("ISOCORRECTOR LOG-FILE - ", logEnvironment$param$Timestamp, "\nIsoCorrectoR-version: ", logEnvironment$param$version, "\n\n", disclaimer, 
               "\n"), file = logFile)
  
  # ERRORS AND WARNINGS
  
  if (length(logEnvironment$error) > 0) {
    
    write(paste0("AN ERROR HAS OCCURED, THE CORRECTION PROCESS WAS ABORTED. ERROR:\n\n", paste(logEnvironment$error, collapse = "\nError in function: "), 
                 "\n"), file = logFile, append = TRUE)
    
  }
  
  if (length(logEnvironment$warning) > 0) {
    
    write(paste0("WARNINGS HAVE OCCURRED, THE CORRECTION MAY BE FAULTY. WARNING(s):\n\n", paste(logEnvironment$warning, collapse = "\n\n"), "\n"), file = logFile, 
          append = TRUE)
    
  }
  
  # FILES AND DIRECTORIES
  
  write(paste0("FILES AND DIRECTORIES\n\nMeasurement File: ", logEnvironment$param$MeasurementFile, "\nMolecule File: ", logEnvironment$param$MoleculeFile, 
               "\nElement File: ", logEnvironment$param$ElementFile, "\nWorking Directory: ", logEnvironment$param$WorkingDirectory, "\nOutput Directory: ", logEnvironment$param$OutputDirectory, 
               "\nResult File(s): ", logEnvironment$param$FileOut, "\n"), file = logFile, append = TRUE)
  
  # PARAMETERS
  
  write(paste0("PARAMETERS\n\nCorrection for tracer impurity: ", logEnvironment$param$CorrectTracerImpurity, "\nCorrection of tracer element in core molecule: ", 
               logEnvironment$param$CorrectTracerElementCore, "\nUltra-High-Resolution correction mode: ", logEnvironment$param$UltraHighRes, "\nCalculation of mean enrichment: ", 
               logEnvironment$param$CalculateMeanEnrichment, "\nMolecules in the correction process: ", paste(logEnvironment$param$molecules, collapse = ", "), "\nTracer(s): ", 
               paste(names(logEnvironment$param$tracers), collapse = ", "), "\nTracer purity: ", paste(logEnvironment$param$tracers, collapse = ", "), "\nLimit value for calculation: ", 
               logEnvironment$param$threshold, "\n"), file = logFile, append = TRUE)
  
}

# Function that is called in the case of anticipated exceptions or warnings for a clean exit and logging of warnings

errorHandler <- function(notification, logEnvironment, type, verbose) {
  
  if (type == "warning" || type == "general") {
    
    logEnvironment[[type]] <- c(logEnvironment[[type]], notification)
    
    if (type == "warning") {
      
      warning(notification)
      
    } else if (type == "general") {
      
      if(verbose){message(date(), " ", notification)}
      
    }
  } else if (type == "error") {
    
    stop(notification)
    
  } else {
    
    notification <- paste0("Invalid error-type ", type, " supplied to errorHandler function")
    
    stop(notification)
    
  }
  
}

# Function to check the structure of measurement data

checkRawData <- function(inputFile, logEnvironment, verbose) {
  
  if(verbose){message(date(), " :: checking raw data ...")}
  
  location <- "In measurement data file: "
  
  data <- as.data.frame(readFileByExt(inputFile, verbose=verbose))
  
  if (ncol(data) < 2) {
    
    notification <- paste0(location, ncol(data), " columns detected but at least 2 are required.")
    errorHandler(notification, logEnvironment, "error", verbose=verbose)
    
  }
  
  if (nrow(data) < 2) {
    
    notification <- paste0(location, nrow(data), " rows of measurements detected but at least 2 are required.")
    errorHandler(notification, logEnvironment, "error", verbose=verbose)
    
  }
  
  if (names(data[1]) != "Measurements/Samples") {
    
    notification <- paste0(location, "Entry in column 1, row 1 must be 'Measurements/Samples'.")
    errorHandler(notification, logEnvironment, "error", verbose=verbose)
    
  }
  
  duplicateRow <- anyDuplicated(data[[1]])
  
  if (duplicateRow) {
    
    notification <- paste0(location, "Duplicate measurement identifier in row ", duplicateRow, " of column 1.")
    errorHandler(notification, logEnvironment, "error", verbose=verbose)
    
  }
  
  duplicateCol <- anyDuplicated(colnames(data))
  
  if (duplicateCol) {
    
    notification <- paste0(location, "Duplicate sample name in column ", duplicateCol, ".")
    errorHandler(notification, logEnvironment, "error", verbose=verbose)
    
  }
  
  rawDataMoleculeNames <- as.character(data[, 1])
  
  checkRes <- checkRawDataMoleculeNames(rawDataMoleculeNames, splitString = "_")
  
  MoleculesFound <- checkRes$moleculesFound
  NoMoleculeSyntax <- checkRes$invalidMolecules
  
  # there are some invalid molecules
  if (length(NoMoleculeSyntax) != 0) {
    
    notification <- stringr::str_c(location, paste0(length(NoMoleculeSyntax), " measurement identifiers in column 1 of the measurement data did not show proper syntax (name of molecule followed by _): ", 
                                                    paste0(NoMoleculeSyntax, collapse = ", ")))
    errorHandler(notification, logEnvironment, "error", verbose=verbose)
  }
  
  for (col in 2:ncol(data)) {
    
    #if (is.element(FALSE, sapply(data[[col]], function(x) (is.numeric(x) || is.na(x))))) {
    if (is.element(FALSE, vapply(data[[col]], function(x) (is.numeric(x) || is.na(x)),TRUE))) {
      notification <- stringr::str_c(location, "Invalid value (not numeric and not NA) in column ", col, " of measurement file.")
      errorHandler(notification, logEnvironment, "error", verbose=verbose)
      
    }
    
  }
  
  MoleculesFound <- unique(MoleculesFound)
  
  if(verbose){message(date(), " :: checking raw data [OK]\n")}
  
  return(list(RawData = data, MoleculesFound = MoleculesFound))
  
}

checkRawDataMoleculeNames <- function(moleculeNames, splitString) {
  tmp.split <- strsplit(moleculeNames, split = splitString)
  wrongSyntax.index <- which(lapply(tmp.split, function(x) length(x) == 1) == TRUE)  #no split
  if (length(wrongSyntax.index) != 0) {
    tmp.split <- tmp.split[-wrongSyntax.index]
  }
  
  return(list(moleculesFound = unlist(lapply(lapply(tmp.split, function(x) x[seq_len(length(x) - 1)]), function(x) if (length(x) == 1) {
    x
  } else {
    paste0(x, collapse = splitString)
  })), invalidMolecules = moleculeNames[wrongSyntax.index]))
}

# Function to check the structure of the supplied molecule file

checkMoleculeDataStructure <- function(data, logEnvironment, verbose) {
  if(verbose){message(date(), " :: checking molecule data structure ...")}
  
  # Check number of columns
  
  if (ncol(data) != 3) {
    
    notification <- paste0(ncol(data), " columns detected in Molecule File but 3 are required.")
    errorHandler(notification, logEnvironment, "error", verbose=verbose)
    
  }
  
  # Check column names
  
  colnamesMoleculeFileExp <- c("molecule", "ms ion or ms/ms product ion", "ms/ms neutral loss")
  colnamesMoleculeFile <- tolower(colnames(data))
  
  for (col in seq_len(ncol(data))) {
    
    if (colnamesMoleculeFileExp[col] != colnamesMoleculeFile[col]) {
      
      notification <- paste0("Column names '", paste0(colnamesMoleculeFile, collapse = ", "), "' found in the molecule file do not match those expected for the file: '", 
                             paste0(colnamesMoleculeFileExp, collapse = ", "), "'.")
      errorHandler(notification, logEnvironment, "error", verbose=verbose)
      
    }
  }
  
  # Check if there are rows
  
  if (nrow(data) == 0) {
    
    notification <- "No rows found in molecule file."
    errorHandler(notification, logEnvironment, "error", verbose=verbose)
    
  }
  
  # Use regular expressions to check column data
  
  moleculeRegex <- "^((|Lab)[A-Z][a-z]?[1-9][0-9]*)+$"
  
  invalidSyntax <- " doesn't match the required syntax."
  
  for (row in seq_len(nrow(data))) {
    
    moleculeName <- data[[1]][[row]]
    
    # Check if Molecule names are NA
    
    if (is.na(moleculeName)) {
      
      notification <- paste0("In molecule file row ", row, ": No molecule name specified.")
      errorHandler(notification, logEnvironment, "error", verbose=verbose)
      
    }
    
    productIon <- data[[2]][[row]]
    neutralLoss <- data[[3]][[row]]
    
    substr1 <- "In molecule file column "
    substr2 <- paste0(", row ", row, ", molecule ", moleculeName, ": Molecule formula ")
    
    locationSpecifierProduct <- paste0(substr1, "2", substr2)
    locationSpecifierNeutral <- paste0(substr1, "3", substr2)
    
    if (is.na(productIon) == FALSE && is.na(neutralLoss) == FALSE) {
      
      if (stringr::str_detect(productIon, moleculeRegex) == FALSE) {
        
        notification <- paste0(locationSpecifierProduct, productIon, invalidSyntax)
        errorHandler(notification, logEnvironment, "error", verbose=verbose)
        
      } else if (stringr::str_detect(neutralLoss, moleculeRegex) == FALSE) {
        
        notification <- paste0(locationSpecifierNeutral, neutralLoss, invalidSyntax)
        errorHandler(notification, logEnvironment, "error", verbose=verbose)
        
      }
      
    } else if (is.na(productIon) == FALSE) {
      
      if (stringr::str_detect(productIon, moleculeRegex) == FALSE) {
        
        notification <- paste0(locationSpecifierProduct, productIon, invalidSyntax)
        errorHandler(notification, logEnvironment, "error", verbose=verbose)
        
      }
    } else {
      
      notification <- paste0(locationSpecifierProduct, "for ", colnamesMoleculeFileExp[2], " must be provided.")
      errorHandler(notification, logEnvironment, "error", verbose=verbose)
      
    }
    
    duplicate <- anyDuplicated(data[[1]])
    
    if (duplicate != 0) {
      
      notification <- paste0("In molecule file: Molecule ", data[[1]][duplicate], " present in duplicate.")
      errorHandler(notification, logEnvironment, "error", verbose=verbose)
      
    }
  }
  if(verbose){message(date(), " :: checking molecule data structure [OK]\n")}
  
  return(data.frame(data))
}

# Check if molecule data makes sense

checkMoleculeDataLogic <- function(extractedData, UltraHighRes, CorrectTracerImpurity, ElementList, logEnvironment, verbose) {
  if(verbose){message(date(), " :: :: checking molecule data logic ...")}
  molecules <- names(extractedData)
  
  numberFragments <- list()
  
  tracers <- list()
  
  for (molecule in molecules) {
    
    numberFragments[[molecule]] <- sum(stringr::str_detect(names(extractedData[[molecule]]), "Fragment"))
    
    tracers[[molecule]] <- c()
    
    tracers[[molecule]][1] <- length(extractedData[[molecule]]$Fragment_1$Tracer)
    
    if (numberFragments[[molecule]] > 1) {
      
      tracers[[molecule]][2] <- length(extractedData[[molecule]]$Fragment_2$Tracer)
      
    } else {
      
      tracers[[molecule]][2] <- 0
      
    }
    
    if (sum(tracers[[molecule]]) == 0) {
      
      notification <- paste0("In molecule file: No tracer element provided for molecule ", molecule, ".")
      errorHandler(notification, logEnvironment, "error", verbose=verbose)
      
    }
    
  }
  
  if (UltraHighRes == TRUE) {
    
    for (molecule in molecules) {
      
      if (numberFragments[[molecule]] != 1) {
        
        notification <- paste0("In molecule file: More than one fragment provided for molecule ", molecule, ", but ultra high resolution correction is not compatible with MS/MS data.")
        errorHandler(notification, logEnvironment, "error", verbose=verbose)
        
      }
      
    }
    
  } else {
    
    for (molecule in molecules) {
      
      if ((tracers[[molecule]][1] > 1) || (tracers[[molecule]][2] > 1)) {
        
        notification <- paste0("In molecule file: More than one tracer element provided for ms ion/product Ion and/or neutral loss of molecule ", molecule, ", but normal resolution correction is only compatible with a single tracer element.")
        errorHandler(notification, logEnvironment, "error", verbose=verbose)
        
      }
      
      if ((tracers[[molecule]][1] > 0) && (tracers[[molecule]][2] > 0)) {
        
        if (extractedData[[molecule]]$Fragment_1$IDTracer != extractedData[[molecule]]$Fragment_2$IDTracer) {
          
          notification <- paste0("In molecule file: Tracer elements in product ion and neutral loss of molecule ", molecule, " do not match.")
          errorHandler(notification, logEnvironment, "error", verbose=verbose)
          
        }
      }
    }
  }
  
  # Check for duplicate elements in the molecule formula
  
  for (molecule in molecules) {
    
    duplicatesPresent <- 0
    
    for (fragmentNo in seq_len(numberFragments[[molecule]])) {
      
      fragment <- paste0("Fragment_", fragmentNo)
      
      duplicatesPresent <- duplicatesPresent + anyDuplicated(names(extractedData[[molecule]][[fragment]]$Element)) + anyDuplicated(names(extractedData[[molecule]][[fragment]]$Tracer))
      
    }
    
    if (duplicatesPresent) {
      
      notification <- paste0("In molecule file: Molecule ", molecule, " contains duplicate elements.")
      errorHandler(notification, logEnvironment, "error", verbose=verbose)
      
    }
    
  }
  
  elementsProvided <- names(ElementList)
  
  for (molecule in molecules) {
    
    for (fragmentNo in seq_len(numberFragments[[molecule]])) {
      
      fragment <- paste0("Fragment_", fragmentNo)
      
      elementsFragment <- extractedData[[molecule]][[fragment]]$Element
      
      if (fragmentNo == 1) {
        fragmentString = "ms ion/product ion"
      } else {
        fragmentString = "neutral loss"
      }
      
      commonString <- paste0("In molecule file: In ", fragmentString, " of molecule ", molecule, ": ")
      
      # Check if elements in the molecule formula are present in the element file
      
      if (anyNA(match(names(elementsFragment), elementsProvided))) {
          
        notification <- paste0(commonString, "One or more elements in the provided formula were not found in the element file.")
        errorHandler(notification, logEnvironment, "error", verbose=verbose)
          
      }
  
      # Do this only if the fragment contains a tracer
      
      if (tracers[[molecule]][[fragmentNo]] > 0) {
        
        tracersFragment <- extractedData[[molecule]][[fragment]]$Tracer
        
        # Check if tracer elements in the molecule formula are also present as non-tracer elements (which gives the total number of atoms, not just the tracer
        # atoms)
        
        if (anyNA(match(names(tracersFragment), names(elementsFragment)))) {
            
          notification <- paste0(commonString, "Information on total atom count (labelled + unlabelled) missing for tracer element(s).")
          errorHandler(notification, logEnvironment, "error", verbose=verbose)
            
        }
          
        # For each tracer element found in the molecule formula, check the number of tracer atoms exceeds the total number of atoms of that element
        
        for (tracer in names(tracersFragment)) {
          
          if (tracersFragment[[tracer]] > elementsFragment[[tracer]]) {
            
            notification <- paste0(commonString, "Number of tracer isotopes of element ", tracer, " exceeds number of atoms of that element present in total.")
            errorHandler(notification, logEnvironment, "error")
            
          }
          
          # For each tracer element, check if a mass shift is provided in the element file
          
          TracerMassShift <- ElementList[[tracer]]$Tracer.isotope.mass.shift
          
          if (TracerMassShift == 0) {
            
            notification <- paste0(commonString, "No mass shift information for tracer element ", tracer, " provided in element file.")
            errorHandler(notification, logEnvironment, "error", verbose=verbose)
            
          }
          
        }
          
        # Given correction for tracer impurity is active, for each tracer element, check if purity is provided in the element file
          
        if (CorrectTracerImpurity == TRUE) {
          
          for (tracer in names(tracersFragment)) {
            
            if (ElementList[[tracer]]$Tracer.purity == 0) {
              
              notification <- paste0(commonString, "No tracer purity information for tracer element ", tracer, " provided in element file, but correction for tracer impurity is active.")
              errorHandler(notification, logEnvironment, "error", verbose=verbose)
              
            }
            
          }
          
        }
          
      } 
      
    }
    
  }
  if(verbose){message(date(), " :: :: checking molecule data logic [OK]")}
}

checkElementDataStructure <- function(data, logEnvironment, verbose) {
  if(verbose){message(date(), " :: checking element data structure ...")}
  # Check number of columns
  
  if (ncol(data) != 4) {
    
    notification <- paste0(ncol(data), " columns detected in Element File but 4 are required.")
    errorHandler(notification, logEnvironment, "error", verbose=verbose)
    
  }
  
  # Check column names
  
  colnamesElementFileExp <- c("element", "isotope abundance_mass shift", "tracer isotope mass shift", "tracer purity")
  colnamesElementFile <- tolower(colnames(data))
  
  for (col in seq_len(ncol(data))) {
    
    if (colnamesElementFileExp[col] != colnamesElementFile[col]) {
      
      notification <- paste0("Column names '", paste0(colnamesElementFile, collapse = ", "), "' found in the element file do not match those expected for the file: '", 
                             paste0(colnamesElementFileExp, collapse = ", "), "'.")
      errorHandler(notification, logEnvironment, "error", verbose=verbose)
      
    }
    
  }
  
  # Check if there are rows
  
  if (nrow(data) == 0) {
    
    notification <- "No rows found in element file."
    errorHandler(notification, logEnvironment, "error", verbose=verbose)
    
  }
  
  # Use regular expressions to check column data
  
  regexVector <- c()
  
  regexVector[1] <- "^[A-Z][a-z]?$"
  
  regexVector[2] <- "(((^1|^1\\.[0]+|^[0]\\.[0-9]+)_(|-)[0-9])(/((1|1\\.[0]+|[0]\\.[0-9]+)_(|-)[0-9]))*)$"
  
  regexVector[3] <- "^(|-)[1-9]$"
  regexVector[4] <- "^1$|^1\\.[0]+$|^[0]\\.[0-9]+$"
  
  notificationVector <- c()
  
  notificationVector[1] <- " is no valid element symbol."
  notificationVector[2] <- " doesn't show appropriate isotope abundance/mass shift definition syntax."
  notificationVector[3] <- " is no valid tracer isotope mass shift definition."
  notificationVector[4] <- " is no valid tracer purity definition."
  
  for (col in seq_len(ncol(data))) {
    
    currentCol <- data[[col]]
    
    # This conversion of NA to empty string for columns 1 and 2 is done because NA values in columns 3 and 4 must not throw an error
    
    for (row in seq_len(nrow(data))) {
      
      if (is.na(currentCol[row]) && col == (1 || 2)) {
        
        currentCol[row] <- ""
        
      }
      
      if (is.na(currentCol[row]) == FALSE) {
        
        if (stringr::str_detect(currentCol[row], regexVector[col]) == FALSE) {
          
          locationSpecifier <- paste0("In element file column ", col, ", row ", row, ": ")
          
          notification <- paste0(locationSpecifier, currentCol[row], notificationVector[col])
          
          errorHandler(notification, logEnvironment, "error", verbose=verbose)
          
        }
      }
    }
  }
  
  # Check for duplicate elements
  
  duplicate <- anyDuplicated(data[[1]])
  
  if (duplicate != 0) {
    
    notification <- paste0("In element file: Element ", data[[1]][duplicate], " present in duplicate.")
    errorHandler(notification, logEnvironment, "error", verbose=verbose)
    
  }
  
  if(verbose){message(date(), " :: checking element data structure [OK]\n")}
  
  return(data.frame(data))
}

# Check if element data makes sense

checkElementDataLogic <- function(extractedData, UltraHighRes, logEnvironment, verbose) {
  #if(verbose){message(date(), " [checkElementDataLogic]")}
  if(verbose){message(date(), " :: :: checking element data logic ...")}
  for (element in names(extractedData)) {
    
    if (any(duplicated(as.data.frame(extractedData[[element]]$Isotopes)$MassShift))) {
      notification <- paste0("In element file: Duplicated MassShift for element ", element, ".")
      errorHandler(notification, logEnvironment, "error")
    }
    
    
    # Without rounding, Si doesn't work as the sum is 1.000001. Howver, should we really round here?
    if (round(sum(data.frame(extractedData[[element]][["Isotopes"]])[, 1]), 5) > 1) {
      
      notification <- paste0("In element file: Sum of isotope abundances of element ", element, " is > 1.")
      errorHandler(notification, logEnvironment, "error", verbose=verbose)
      
    }
    
  }
  
  # In case of UHR mode, check if the mass shift of the tracers is also found in the tracer elements natural isotopes. If not, probabilities cannot be
  # calculated
  
  if (UltraHighRes) {
    
    for (element in names(extractedData)) {
      
      TracerMassShift <- extractedData[[element]]$Tracer.isotope.mass.shift
      
      if (TracerMassShift > 0 && is.na(match(TracerMassShift, as.data.frame(extractedData[[element]]$Isotopes)$MassShift))) {
        
        notification <- paste0("In element file: Mass shift '", TracerMassShift, "' of tracer element ", element, " is not found in the natural isotope information of that element.")
        errorHandler(notification, logEnvironment, "error", verbose=verbose)
        
      }
      
    }
    
  }
  
  if(verbose){message(date(), " :: :: checking element data logic [OK]")}
  
}

readFileByExt <- function(file, verbose) {

  fileExt <- tools::file_ext(file)
  if (fileExt == "") {
    stop(date(), " :: no file extension for file: ", file, "\n")
  } else {
    if(verbose){message(date(), " :: :: detected file format: ", fileExt)}
    if (fileExt %in% c("xls", "xlsx")) {
      fileData <- readxl::read_excel(file)
    } else if (fileExt == "csv") {
      fileData <- readr::read_csv(file)
    } else {
      stop(date(), " :: unsupported file format: ", fileExt, "\n")
    }
  }
  return(fileData)
}

convertEnrichmentList2matrix <- function(input, verbose) {
  if(verbose){message(date(), " :: :: converting enrichment results into a matrix ...")}
  # matrix with sampleIds as columns and mean enrichment for molecule on row
  m <- matrix(ncol = length(input), nrow = length(input[[1]]))
  colnames(m) <- names(input)
  rownames(m) <- names(input[[1]])
  
  # TODO: quick and dirty loop. find more elegant solution !!
  for (i in seq_len(length(input))) {
    for (i2 in seq_len(length(input[[i]]))) {
      m[i2, i] <- as.numeric(input[[i]][[i2]])
    }  #i2
  }  #i
  if(verbose){message(date(), " :: :: converting enrichment results into a matrix [OK]")}
  return(m)
  
}


convert2matrix <- function(input) {
  vec.transitions <- vector()
  # we need to loop over all samples to really get all transitions
  for (sample in names(input)) {
    for (molecule in names(input[[sample]])) {
      vec.transitions <- c(vec.transitions, names(input[[sample]][[molecule]][[1]]))
    }  #molecule
  }  #sample
  transitions.unique <- unique(vec.transitions)
  
  templateMat <- matrix(NA, ncol = length(names(input)), nrow = length(transitions.unique))
  colnames(templateMat) <- names(input)
  rownames(templateMat) <- transitions.unique
  
  # now populate the four matrices
  list.corrMat <- list()
  for (corrMat in names(input[[1]][[1]])) {
    
    templateMat2 <- templateMat
    for (sample in names(input)) {
      for (molecule in names(input[[sample]])) {
        for (transition in names(input[[sample]][[molecule]][[corrMat]])) {
          templateMat2[transition, sample] <- input[[sample]][[molecule]][[corrMat]][[transition]]
        }  #transition
      }  #molecule
    }  #sample
    list.corrMat[[corrMat]] <- templateMat2
  }  #corrMat
  
  return(list.corrMat)
}  #convert2matrix

defineOutputFilenames <- function(logEnvironment,FileOut,FileOutFormat,
                                  CalculateMeanEnrichment,CorrectAlsoMonoisotopic) {
  
  #DEFINE NAMES OF OUTPUT FILES
  
  logEnvironment$param$LogfileName <-
    paste0(logEnvironment$param$NamePrefix, "_", FileOut, ".log")
  
  CorrectedDataNamesBase <-
    c("Corrected",
      "CorrectedFractions",
      "Residuals",
      "RelativeResiduals")
  
  if (CorrectAlsoMonoisotopic) {
    CorrectedDataNamesBase <-
      c(
        CorrectedDataNamesBase,
        "CorrectedMonoisotopic",
        "CorrectedMonoisotopicFractions"
      )
    
  }
  
  if (CalculateMeanEnrichment) {
    CorrectedDataNames <-
      c(CorrectedDataNamesBase, "MeanEnrichment", "RawData")
    
  }
  else {
    CorrectedDataNames <- c(CorrectedDataNamesBase, "RawData")
    
  }
  
  if (FileOutFormat == "csv") {
    CorrectedDataFileNames <-
      vector(mode = "character",
             length = length(CorrectedDataNames))
    names(CorrectedDataFileNames) <- CorrectedDataNames
    
    for (name in CorrectedDataNames) {
      CorrectedDataFileNames[name] <-
        paste0(logEnvironment$param$NamePrefix,
               "_",
               FileOut,
               "_",
               name,
               ".",
               FileOutFormat)
      
    }
    
    logEnvironment$param$FileOut <-
      paste0(CorrectedDataFileNames, collapse = ", ")
    
  }
  else if (FileOutFormat == "xls") {
    logEnvironment$param$FileOut <-
      paste0(logEnvironment$param$NamePrefix,
             "_",
             FileOut,
             ".",
             FileOutFormat)
    
    CorrectedDataFileNames <- NA
    
  }
  
  return(list("CorrectedDataNames" = CorrectedDataNames,
              "CorrectedDataNamesBase" = CorrectedDataNamesBase,
              "CorrectedDataFileNames" = CorrectedDataFileNames))
  
}

#WRITE CORRECTED DATA TO FILE

writeCorrectionResultsToFile <- function(CorrectedDataOutput,NamesCorrectedOutput,logEnvironment, verbose) {

  if(verbose){message(date()," :: writing results to ",logEnvironment$param$FileOut," ...")}
  if (logEnvironment$param$FileOutFormat == "xls") {
    WriteXLS::WriteXLS(
      CorrectedDataOutput,
      file.path(logEnvironment$param$OutputDirectory, logEnvironment$param$FileOut),
      perl = "perl",
      row.names = TRUE
    )
    
  } else if (logEnvironment$param$FileOutFormat == "csv") {
    for (name in NamesCorrectedOutput$CorrectedDataNames) {
      utils::write.csv(
        CorrectedDataOutput[[name]],
        file = file.path(
          logEnvironment$param$OutputDirectory,
          NamesCorrectedOutput$CorrectedDataFileNames[name]
        )
      )
      
    }
  }
  if(verbose){message(date()," :: writing results to ",logEnvironment$param$FileOut," [OK]\n")}
}

# Data correction using the probability matrix

# The function 'RawDataCorrection' performs the actual
# data correction. It uses the values extracted from the measurement file
# and corrects molecule by molecule for each sample. This is done by
# numerically solving a linear equation system. Here,
# the uncorrected value of a given transition is a linear combination
# of the corrected transition values with their
# probability of contributing to the uncorrected value
# as their coefficients (derived from the probability matrix,
# ProbMatrixComplete). The algorithm works with the
# constraint that a solution of the linear equation
# system cannot be < 0.

#' @importFrom pracma lsqlincon
RawDataCorrection <- function(UncorrectedData, MoleculeData, MoleculeName,
                              ProbMatrix, MoleculeNo, SampleNo, SampleName,
                              roundDigits, logEnvironment, verbose) {
  NumberTransitions <- nrow(MoleculeData[["Transitions"]])

  ProbMatrixComplete <- ProbMatrix

  # In the following, the linear equation system is solved with linear
  # inequality constraints: ConstraintMatrix*SolutionVector <= ConstraintVector.
  # The ConstraintVector is 0, the ConstraintMatrix is -1 at each
  # ConstraintMatrix(TransitionNo, TransitionNo) position.
  # It is thereby assured that the solutions cannot be < 0.
  # If a solution is < 0, ConstraintMatrix*SolutionVector becomes positive and
  # the constraint ConstraintMatrix*SolutionVector <=
  # ConstraintVector (which is 0) is not fulfilled.

  ConstraintVector <- rep(0, NumberTransitions)
  ConstraintMatrix <- matrix(0, nrow = NumberTransitions, ncol = NumberTransitions)

  for (TransitionNo in seq_len(NumberTransitions)) {
    ConstraintMatrix[TransitionNo, TransitionNo] <- -1
  }

  # Check the vector of uncorrected values for values that are not a number
  # (e.g. NA) and thus missing.
  CheckForNaN <- is.na(UncorrectedData)

  # If there are no missing values, perform the correction for the
  # current vector of uncorrected values UncorrectedData.
  # This is done by numerically solving a linear equation system where each
  # uncorrected value is a linear combination of unknown corrected values with
  # their probabilites of contributing to the uncorrected value as the constants
  # (from ProbMatrix).  ||ProbMatrix*CorrTransitionsPlus - UncorrectedData||^2 is minimized in a linear least squares
  # approach with the constraint that the corrected values (CorrTransitionsPlus)
  # must not be < 0. In addition to CorrTransitionsPlus, also the residuals of
  # the solving process are given (CorrResiduals).
  
  if (sum(CheckForNaN) == 0) {
    CorrTransitionsPlus <- pracma::lsqlincon(C = ProbMatrix, d = as.numeric(UncorrectedData), A = ConstraintMatrix, b = ConstraintVector)
    ManResiduals <- ProbMatrix %*% CorrTransitionsPlus - UncorrectedData
  } else {
    notification <- paste0(
      "In measurement data file: Measurement data of molecule ", MoleculeName, " in sample ", SampleName, " contains NA values. The correction performed may be less acurate.",
      "\nBe especially careful when considering fraction and mean enrichment values from samples with missing values."
    )
    errorHandler(notification, logEnvironment, "warning", verbose = verbose)

    CorrTransitionsPlus <- vector()
    ManResiduals <- vector()

    # Get indices of missing values. The rows with these indices removed from
    # UncorrectedData and the ConstraintVector. Row and column with these
    # indices are removed from ProbMatrix and ConstraintMatrix.
    # This way, the missing values are completely removed from the
    # linear equation system.

    # get index of NaN values
    nan.index <- as.numeric(which(CheckForNaN))

    # remove respective rows/columns
    UncorrectedData.nan <- UncorrectedData[-nan.index]
    ProbMatrix.nan <- ProbMatrix[-nan.index, -nan.index]
    ConstraintMatrix.nan <- ConstraintMatrix[-nan.index, -nan.index]
    ConstraintVector.nan <- ConstraintVector[-nan.index]

    # Here, the linear equation system is solved with a set of equations that is reduced because of the missing values. If the missing values are expected to
    # be high, the effect on the accuracy of the correction is more pronounced than if they are expected to be close to 0.

    if (length(nan.index) < NumberTransitions) {
      CorrTransitionsPlusNaN <- pracma::lsqlincon(C = ProbMatrix.nan, d = as.numeric(UncorrectedData.nan), A = ConstraintMatrix.nan, b = ConstraintVector.nan)
      ManResidualsNaN <- ProbMatrix.nan %*% CorrTransitionsPlusNaN - UncorrectedData.nan
    } else {
      notification <- paste0(
        "In measurement data file: All measurements are NA for molecule ", MoleculeName, " in sample ", SampleName ,
        "."
      )
      errorHandler(notification, logEnvironment, "warning", verbose)
    }

    # To be able to correctly assign corrected values to their measurement tag when missing values are present, a full size corrected value vector
    # CorrTransitionsPlus is made from the reduced size corrected values vector CorrTransitionsPlusNaN.  This is done by adding the missing values as NA at
    # the associated vector index.  In the same way a full size CorrResiduals vector is produced. A warning is given in the WarningsLog for each missing
    # value.

    i <- 1

    # NumberTransitions<-ncol(ProbMatrix)

    for (TransitionNo in seq_len(NumberTransitions)) {
      if (!CheckForNaN[TransitionNo]) {
        # 2017-08-28

        CorrTransitionsPlus[TransitionNo] <- CorrTransitionsPlusNaN[i]
        ManResiduals[TransitionNo] <- ManResidualsNaN[i]

        i <- i + 1
      } else {
        CorrTransitionsPlus[TransitionNo] <- NA
        ManResiduals[TransitionNo] <- NA
      } # if
    } # TransitionNo
  } # sum(CheckForNaN)==0

  CorrTransitionsPlus <- round(CorrTransitionsPlus, digits = roundDigits)
  ManResiduals <- round(ManResiduals, digits = roundDigits)

  # The corrected values in CorrTransitionsPlus are values that correspond to
  # the full integral of the isotopologue abundance distribution of a given
  # labeled species. To get the value (CorrTransitions) that corresponds to the
  # species in the distribution that contains no natural isotopes of higher
  # mass, the corresponding CorrTransitionsPlus value has to be multiplied
  # with the probability that the species contains no isotopes of higher mass
  # due to natural abundance

  CorrTransitions <- vector()
  CorrTransitionsFractions <- vector()
  CorrTransitionsPlusFractions <- vector()
  RelativeResiduals <- vector()

  # Calculation of residuals relative to corrected data

  RelativeResiduals <- round(ManResiduals / CorrTransitionsPlus, digits = roundDigits)

  for (TransitionNo in seq_len(NumberTransitions)) {
    if (is.na(CorrTransitionsPlus[TransitionNo]) == FALSE) {
      CorrTransitions[TransitionNo] <- CorrTransitionsPlus[TransitionNo] * ProbMatrixComplete[TransitionNo, TransitionNo]
    } else {
      CorrTransitions[TransitionNo] <- NA
    }
  } # TransitionNo

  round(CorrTransitions, digits = roundDigits)

  # To be able to calculate fractions, the vectors of the corrected values are summed.
  CorrTransitionsPlusSum <- sum(CorrTransitionsPlus[which(CheckForNaN == FALSE)])
  CorrTransitionsSum <- sum(CorrTransitions[which(CheckForNaN == FALSE)])

  # Calculation of fractions

  CorrTransitionsFractions <- vector()
  CorrTransitionsPlusFractions <- vector()

  for (TransitionNo in seq_len(NumberTransitions)) {
    CorrTransitionsFractions[TransitionNo] <- CorrTransitions[TransitionNo] / CorrTransitionsSum
    CorrTransitionsPlusFractions[TransitionNo] <- CorrTransitionsPlus[TransitionNo] / CorrTransitionsPlusSum
  } # TransitionNo

  round(CorrTransitionsPlusFractions, digits = roundDigits)
  round(CorrTransitionsFractions, digits = roundDigits)

  names(CorrTransitions) <- colnames(ProbMatrix)
  names(CorrTransitionsPlus) <- colnames(ProbMatrix)
  names(ManResiduals) <- colnames(ProbMatrix)
  names(RelativeResiduals) <- colnames(ProbMatrix)
  names(CorrTransitionsFractions) <- colnames(ProbMatrix)
  names(CorrTransitionsPlusFractions) <- colnames(ProbMatrix)

  returnList <- list(
    CorrectedMonoisotopic = CorrTransitions, Corrected = CorrTransitionsPlus, CorrectedMonoisotopicFractions = CorrTransitionsFractions,
    CorrectedFractions = CorrTransitionsPlusFractions, Residuals = ManResiduals, RelativeResiduals = RelativeResiduals
  )

  return(returnList)
}

# Automated Computation of Transitions

#' @importFrom stringr str_detect
#' @importFrom stringr str_c
#' @importFrom magrittr '%>%'
#' 
  CalculateTransitionsUHR <- function(MoleculeInfo, ElementInfo, verbose) {
  
  if(verbose){message(date(), " :: calculating transitions ...")}
  MoleculesTotal <- length(MoleculeInfo)
  MoleculesName <- names(MoleculeInfo)
    
  # CalculateTransitions() in the case of UltraHighRes option active for n different tracers.
  
  TotalCombinationsList <- list()
  TransitionsExpectedList <- list()
  for (MoleculeNo in seq_len(MoleculesTotal)) {
    
    MoleculeData <- MoleculeInfo[[MoleculeNo]]  # entire 'MoleculeInfo' data for current molecule
    NumberTracers <- unlist(lapply(MoleculeData, function(x) length(x[["Tracer"]])))[[1]]  # number of tracers of First Fragment only !
    
    MaxLabel <- lapply(MoleculeData, function(x) x[["MaxLabel"]])[[1]]
    IDTracer <- lapply(MoleculeData, function(x) x[["IDTracer"]])[[1]]
    
    tmpTotalCombinationsList <- list()
    
    # SubCombinations.vec[TracerNo] contains the number of possible combined labelling states that can arise when combining the labelling of the
    # [1...TracerNo-1] tracer elements that occur in the list of tracer elements prior to the tracer element with number TracerNo.
    
    SubCombinations.vec <- 1
    
    if (NumberTracers > 0) 
    {
      tmpTransitionsExpectedList <- list()
      # number of tracers
      
      for (TracerNo in seq_len(NumberTracers)) {
        SubCombinations.vec[TracerNo + 1] <- c(SubCombinations.vec[TracerNo] * (MaxLabel[TracerNo] + 1))
      }  #TracerNo 
      
      # TotalCombinations is the total number of labelling states that is possible with the set of tracer elements
      
      TotalCombinations <- SubCombinations.vec[NumberTracers[[1]] + 1]
      
      # Now for each expected labelling state the number of each tracer element is calculated
      
      TransitionsExpected.vec <- vector()
      for (TracerNo in seq_len(NumberTracers)) {
        
        LabelSpacer <- 1
        
        # In this loop, each number of label of a tracer is repeated 
        # SubCombinations.vec[TracerNo] times in succession in the 
        # TransitionsExpected.vec. LabelSpacer
        # then provides the first index of the next number of label in the vector
        
        for (Label in 0:MaxLabel[TracerNo]) {
          
          TransitionsExpected.vec[LabelSpacer:(LabelSpacer + SubCombinations.vec[TracerNo] - 1)] <- Label
          LabelSpacer <- LabelSpacer + SubCombinations.vec[TracerNo]
          
        }  #Label
        
        tmpTransitionsExpectedList[[TracerNo]] <- TransitionsExpected.vec
        
        RepeatSpacer <- LabelSpacer
        
        # Now, the block generated in TransitionsExpected.vec in the previous
        # loop is copied until the length of the vector matches 
        # TotalCombinations. Then, the
        # labelling with the given tracer is present in 
        # TransitionsExpected.vec for all labelling states of the molecule.
        
        for (Repeat in seq_len((TotalCombinations/SubCombinations.vec[TracerNo + 1])) - 1) {
          
          if (Repeat > 0) {
            TransitionsExpected.vec[RepeatSpacer:(RepeatSpacer + SubCombinations.vec[TracerNo + 1] - 1)] <- tmpTransitionsExpectedList[[TracerNo]][seq_len(SubCombinations.vec[TracerNo + 1])]
            
            RepeatSpacer <- RepeatSpacer + SubCombinations.vec[TracerNo + 1]
          }
        }  #Repeat
        
        # This procedure is performed for all tracers to get a list of 
        # expected labelling states
        
        tmpTransitionsExpectedList[[TracerNo]] <- TransitionsExpected.vec
      }  #TracerNo
      
      names(tmpTransitionsExpectedList) <- MoleculeData[[1]][["IDTracer"]]
      
      NumberTransitions <- length(tmpTransitionsExpectedList[[1]])
      
      tmpTransitionsExpectedList[["Sum"]] <- apply(as.data.frame(tmpTransitionsExpectedList), 1, sum)
      
      TransitionsExpected.df <- as.data.frame(tmpTransitionsExpectedList)
      
      # Compute the expected names of the expected labelling states
      
      Prefix <- stringr::str_c(MoleculesName[MoleculeNo], "_")
      
      for (TransitionNo in seq_len(NumberTransitions)) {
        PreName <- Prefix
        
        for (TracerNo in seq_len(NumberTracers)) {
          
          PreName <- stringr::str_c(PreName, IDTracer[TracerNo], as.numeric(TransitionsExpected.df[TransitionNo, TracerNo]))
          if (TracerNo < NumberTracers) 
          {
            PreName <- stringr::str_c(PreName, ".")
          }  #if
          if (TracerNo == NumberTracers) 
          {
            rownames(TransitionsExpected.df)[TransitionNo] <- PreName
          }  #if
          
        }  #TracerNo
        
      }  #TransitionNo
      
      TransitionsExpected.df <- TransitionsExpected.df[order(rownames(TransitionsExpected.df)), ]
      
      MoleculeInfo[[MoleculeNo]][["TransitionsExpected"]] <- TransitionsExpected.df
      MoleculeInfo[[MoleculeNo]][["TotalCombinations"]] <- TotalCombinations
      
    }  #NumberTracers[[1]]>0
    
  }  #MoleculeNo
  
  if(verbose){message(date(), " :: calculating transitions [OK]\n")}
  return(MoleculeInfo)

}

#' @importFrom dplyr last
#' @importFrom dplyr arrange
#' @importFrom stringr str_c
#' @importFrom tibble as_tibble
#' @importFrom magrittr '%>%'

IsoCombinations <- function(Fragment, x, Element, ElementInfo, AvailablePlacesTotal, IsoCluster, CalculationThreshold) {

  # NumberIso is the number of isotopes per element.
  # The number of nested loops used depends on the number of
  # isotopes of the current element.

  NumberIso <- nrow(ElementInfo[[Element]][["Isotopes"]][[1]])

  Combinations <- vector()
  Combinations[1] <- 1

  # Calculation of Combinations[IsotopeNo+1], the number of isotope combinations
  # when considering IsotopeNo different isotopes (also including combinations
  # where the sum of isotopes exceeds the number of available places IsoCluster).

  for (IsotopeNo in seq_len(NumberIso)) {
    Combinations[IsotopeNo + 1] <- Combinations[IsotopeNo] * ((IsoCluster) + 1)
  } # IsotopeNo

  TotalCombinations <- dplyr::last(Combinations) # get last element

  # Computation of the isotope combinations list `CombinationsArrayPreList`.
  # At this point, combinations where the sum of the different isotopes exceeds
  # IsoCluster are still considered, leading to invalid combinations that are
  # later removed. In its columns, CombinationsArrayPreList contains the number
  # of occurences of a given isotope in a given combination.
  # Each row corresponds to a combination.

  CombinationsArrayPreList <- list()

  for (IsotopeNo in seq_len(NumberIso)) {
    p <- 1
    tmpCombinationsArrayPre <- vector()

    # first, for each column (isotope) the repetitive column element is computed

    for (IsoCount in 0:IsoCluster) {
      tmpCombinationsArrayPre[p:(p + Combinations[IsotopeNo] - 1)] <- IsoCount
      p <- p + Combinations[IsotopeNo]
    } # IsoCount

    # Then, the repetitive column element is repeatedly inserted until the
    # number of rows matches `TotalCombinations`

    for (Repeat in seq_len((TotalCombinations / Combinations[IsotopeNo + 1])) - 1) {
      if (Repeat > 0) {
        tmpCombinationsArrayPre[p:(p + Combinations[IsotopeNo + 1] - 1)] <- tmpCombinationsArrayPre[seq_len(Combinations[IsotopeNo + 1])]
        p <- p + Combinations[IsotopeNo + 1]
      }
    } # Repeat

    CombinationsArrayPreList[[IsotopeNo]] <- tmpCombinationsArrayPre
  } # IsotopeNo
  names(CombinationsArrayPreList) <- stringr::str_c("Isotope_", seq_len(NumberIso))

  # CombinationsArrayPreList is sorted to get the most probable combinations on top

  CombinationsArrayPreList2 <- CombinationsArrayPreList %>% as_tibble() %>% dplyr::arrange(pick(names(CombinationsArrayPreList))) %>% as.list()

  CombinationsArrayPre.df <- as.data.frame(lapply(CombinationsArrayPreList2, function(x) rev(x)))

  RowSum <- apply(CombinationsArrayPre.df, 1, sum)
  j <- 1

  # CombinationsArray is derived from CombinationsArrayPreList by removing
  # all combinations that do not have a row sum of IsoCluster

  CombinationsArray <- NULL

  for (i in seq_len(TotalCombinations)) {
    if (RowSum[i] == IsoCluster) {
      tmpA <- vector()

      tmpA <- as.numeric(CombinationsArrayPre.df[i, seq_len(NumberIso)])
      tmpA[1] <- tmpA[1] + (AvailablePlacesTotal - IsoCluster)
      CombinationsArray <- rbind(CombinationsArray, tmpA)

      j <- j + 1
    } # RowSum[i]==IsoCluster
  } # i

  # The number of combinations in CombinationsArray is derived,
  # CombinationsArray is stored in Isotopes

  NumberIsoCombMax <- nrow(CombinationsArray)

  Isotopes <- CombinationsArray

  # The probabilities and mass shifts of the isotope combinations in
  # CombinationsArray are computed using IsoCombinationsProbMassShift()

  IsoComb <- 1
  ProbArray.vec <- vector()
  MassShiftArray.vec <- vector()
  NumberIsoComb.vec <- vector()

  for (IsoCombPre in seq_len(NumberIsoCombMax)) {
    icpms <- IsoCombinationsProbMassShift(
      Element = Element, NumberIso = NumberIso, ElementInfo = ElementInfo, CombinationsArray = CombinationsArray,
      IsoCombPre = IsoCombPre, AvailablePlacesTotal = AvailablePlacesTotal
    )

    # Due to CalculationThreshold which limits calculations to probability
    # values large enough to be relevant, probability and mass shift are
    # usually not computed for all isotope combinations.

    if (icpms[["ProbIsoComb"]] > CalculationThreshold) {
      ProbArray.vec[IsoComb] <- icpms[["ProbIsoComb"]]
      MassShiftArray.vec[IsoComb] <- icpms[["MassShiftIsoComb"]]
      NumberIsoComb <- IsoComb
      IsoComb <- IsoComb + 1
    } else {
      if (IsoComb == 1) {
        NumberIsoComb <- 0
      }

      IsoCombPre <- NumberIsoCombMax + 1
    }
  } # IsoCombPre

  return(list(Isotopes = Isotopes, ProbArray = ProbArray.vec, MassShiftArray = MassShiftArray.vec, NumberIsoComb = NumberIsoComb))
} # IsoCombinations

#' @importFrom utils packageDescription
.onAttach <- function(libname, pkgname) {
    pkgVersion <- packageDescription(pkgname, fields = "Version")
    pkgLicense <- packageDescription(pkgname, fields = "License")
    
    msg <- paste0(pkgname, " v", pkgVersion, "\n\n", "For help, type browseVignettes('IsoCorrectoR') to view the IsoCorrectoR documentation or ?IsoCorrection\n", 
        "to view the documentation of the IsoCorrection() function.\n\n")
    
    disclaimer <- paste0(pkgname, " is licensed under ", pkgLicense, ". It is free software and comes without any warranty.\n\n")
    
    citation <- paste0("If you use ", pkgname, " in published research, please cite:\n\n", "Paul Heinrich, Christian Kohler et al.\n", "Correction for natural isotope abundance and tracer impurity in MS-, MS/MS- and high-resolution multiple-tracer-data from stable isotope labeling experiments with IsoCorrectoR", 
        " (Scientific reports, 2018)\n\n")
    
    packageStartupMessage(paste0(msg, disclaimer, citation))
}

# CALCULATION OF THE TRACER ELEMENT - NON-TRACER ELEMENT COMBINATIONS

# Given that there are more elements than just the tracer element, the element combinations from CombElemProbList are now multiplied with the isotope
# combinations of the tracer element, for all labelling states. This is analogous to the previous algorithms from this function.  However, in the
# computation of the mass shift array, the intrinsic mass shift (due to tracer incorporation) associated with the respective labelling state is added in
# addtion.  In the end, the probability array TracerElemProbList and the mass shift array TracerElemMassShiftList are derived. They contain the
# probabilities and mass shifts of all relevant elemental combinations of isotope combinations for all labelling states of a given molecule(-fragment).

ElementCombinationsTracer <- function(ProbTracerList, MassShiftTracerList, NumberIsoCombTracerEff, 
                                         CombElemProbList, CombElemMassShiftList, NumberEleComb,
                                         Fragment, NumberElementsNonTracer, NumberTracers, Transitions,
                                         NumberTransitions, CalculationThreshold) {

  TracerElemProbList <- list()
  TracerElemMassShiftList <- list()
  NumberTracerElemComb <- vector()
  
  if (NumberTracers > 0) {
    
    if (NumberElementsNonTracer > 0) {
      
      for (TransitionNo in seq_len(NumberTransitions)) {
        
        tmpTracerElemProb_Array <- vector()
        tmpTracerElemMassShift_Array <- vector()
        EleComb <- 1
        
        for (ElementCombination in seq_len(NumberEleComb[[NumberElementsNonTracer]])) {
          
          for (IsoCombination in seq_len(NumberIsoCombTracerEff[[TransitionNo]])) {
            
            TracerElemProb <- CombElemProbList[[NumberElementsNonTracer]][[ElementCombination]] * ProbTracerList[[TransitionNo]][[IsoCombination]]
            if (TracerElemProb > CalculationThreshold) {
              tmpTracerElemProb_Array[EleComb] <- TracerElemProb
              
              tmpTracerElemMassShift_Array[EleComb] <- CombElemMassShiftList[[NumberElementsNonTracer]][ElementCombination] + MassShiftTracerList[[TransitionNo]][IsoCombination] + 
                Transitions[TransitionNo, Fragment]
              
              NumberTracerElemComb[TransitionNo] <- EleComb
              
              EleComb <- EleComb + 1
              
            } else {
              IsoCombination <- NumberIsoCombTracerEff[[TransitionNo]] + 1
            }  #TracerElemProb > CalculationThreshold
          }  #IsoCombination
        }  #ElementCombination
        
        TracerElemProbList[[TransitionNo]] <- tmpTracerElemProb_Array
        TracerElemMassShiftList[[TransitionNo]] <- tmpTracerElemMassShift_Array
        
      }  #TransitionNo
    } else {
      
      # If the only element considered is the tracer element, the probability- and mass shift arrays from 'IsoCombinationsResult' are directly fed into the
      # probability array TracerElemProbList and the mass shift array TracerElemMassShiftList.  However, for each transition its intrinsic mass shift is
      # added to the mass shift array in addition.
      
      for (i in seq_len(NumberTransitions)) {
        NumberTracerElemComb[i] <- NumberIsoCombTracerEff[[i]]
      }  #i
      
      for (TransitionNo in seq_len(NumberTransitions)) {
        tmpTracerElemProb_Array <- vector()
        tmpTracerElemMassShift_Array <- vector()
        
        for (IsoCombination in seq_len(NumberIsoCombTracerEff[[TransitionNo]])) {
          
          tmpTracerElemProb_Array[IsoCombination] <- ProbTracerList[[TransitionNo]][IsoCombination]
          
          tmpTracerElemMassShift_Array[IsoCombination] <- MassShiftTracerList[[TransitionNo]][IsoCombination] + Transitions[TransitionNo, Fragment]
          
        }  #IsoCombination
        TracerElemProbList[[TransitionNo]] <- tmpTracerElemProb_Array
        TracerElemMassShiftList[[TransitionNo]] <- tmpTracerElemMassShift_Array
      }  #TransitionNo
    }  #NumberElementsNonTracer>0
  } else {
    
    # If the molecule(-fragment) in question contains no tracer element, the probability array TracerElemProbList and the mass shift array
    # TracerElemMassShiftList are equal to CombElemProbList and CombElemMassShiftList for all transitions of this molecule(-fragment).
    
    for (TransitionNo in seq_len(NumberTransitions)) {
      
      tmpTracerElemProb_Array <- vector()
      tmpTracerElemMassShift_Array <- vector()
      
      for (i in seq_len(NumberEleComb[[NumberElementsNonTracer]])) {
        
        tmpTracerElemProb_Array[i] <- CombElemProbList[[NumberElementsNonTracer]][i]
        tmpTracerElemMassShift_Array[i] <- CombElemMassShiftList[[NumberElementsNonTracer]][i]
      }  #i
      
      TracerElemProbList[[TransitionNo]] <- tmpTracerElemProb_Array
      TracerElemMassShiftList[[TransitionNo]] <- tmpTracerElemMassShift_Array
      NumberTracerElemComb[TransitionNo] <- NumberEleComb[[NumberElementsNonTracer]]
    }  #TransitionNo
    
  }  #NumberTracers>0
  
  names(TracerElemProbList) <- rownames(Transitions)
  names(TracerElemMassShiftList) <- rownames(Transitions)
  names(NumberTracerElemComb) <- rownames(Transitions)
  
  return(list("TracerElemProbList"=TracerElemProbList,
              "TracerElemMassShiftList"=TracerElemMassShiftList,
              "NumberTracerElemComb"=NumberTracerElemComb))
  
}

library(readxl)

#Variables used by all/many tests

TestDataPath <- system.file('testdata', package='IsoCorrectoR')

resultTypes <- c('c', 'cf', 'me', 'cm', 'rd')
names(resultTypes) <- c('Corrected', 'CorrectedFractions', 'MeanEnrichment',
                        'CorrectedMonoisotopic', 'RawData')

#Safe loading of result files with tryCatch()

loadResults <- function(filepath, filetype='csv', sheet=NA, dataDescription=NA) {
  
  result <- tryCatch(
    {
      
      if(filetype=='csv') {
      
        result <- read.csv(filepath, row.names = 1)
        
      } else if (filetype=='xls') {
        
        result <- as.data.frame(read_excel(filepath, col_names = TRUE, sheet = sheet))
        
        rownames(result) <- result[[1]]
        result <- result[-1]
        
      } else {
        
        stop("Invalid file extension.")
        
      }
      
    },
    error = function(e) {
      
      result <- paste0("No ", dataDescription, " file")
      return(result)
      
    }
  )
  
  return(result)
  
}

#Safe call to IsoCorrection() with tryCatch

IsoCorrectionSafeCall <- function(MeasurementFile=NA,
                                  ElementFile=NA,
                                  MoleculeFile=NA,
                                  CorrectTracerImpurity=FALSE,
                                  CorrectTracerElementCore=TRUE, 
                                  CalculateMeanEnrichment=TRUE,
                                  UltraHighRes=FALSE,
                                  DirOut='.', 
                                  FileOut='result',
                                  FileOutFormat='csv',
                                  CorrectAlsoMonoisotopic=FALSE,
                                  CalculationThreshold=10^(-8),
                                  CalculationThreshold_UHR=8,
                                  ReturnResultsObject=TRUE,
                                  Testmode=TRUE) {

  correctionResults <- tryCatch(
    {
      
      correctionResults <- IsoCorrection(MeasurementFile=MeasurementFile,
                                         ElementFile=ElementFile,
                                         MoleculeFile=MoleculeFile,
                                         CorrectTracerImpurity=CorrectTracerImpurity,
                                         CorrectTracerElementCore=CorrectTracerElementCore, 
                                         CalculateMeanEnrichment=CalculateMeanEnrichment,
                                         UltraHighRes=UltraHighRes,
                                         DirOut=DirOut, 
                                         FileOut=FileOut,
                                         FileOutFormat=FileOutFormat,
                                         CorrectAlsoMonoisotopic=CorrectAlsoMonoisotopic,
                                         CalculationThreshold=CalculationThreshold,
                                         CalculationThreshold_UHR=CalculationThreshold_UHR,
                                         ReturnResultsObject=ReturnResultsObject,
                                         Testmode=Testmode)
      
    }, 
    error = function(e) {
      
      correctionResults <- list()
      
      correctionResults$success <- "UNEXPECTED_ERROR"
      correctionResults$results <- NA
      
      return(correctionResults)
      
    }
  )
}

context("Test correctness of results returned by the IsoCorrection() function")

test_that("Test results returned by IsoCorrection()", {
  
  ExpectedOutcome = 'TRUE'
  
  Project <- 'AA1'
  
  ProjectPath <- file.path(TestDataPath, 'end_to_end_nres', Project)
  
  DirOut = file.path(ProjectPath, 'test_output')
  
  if(!dir.exists(DirOut)) {
    
    dir.create(DirOut)
    
  }
  
  MeasurementFile = file.path(ProjectPath, 'MeasurementFile.csv')
  MoleculeFile = file.path(ProjectPath, 'MoleculeFile.csv')
  ElementFile = file.path(TestDataPath, 'ElementFile.csv')
  
  FileOut = "co_np"
  
  #run IsoCorrection with given parameters
      
  correctionResults <- IsoCorrectionSafeCall(MeasurementFile=MeasurementFile,
                                     ElementFile=ElementFile,
                                     MoleculeFile=MoleculeFile,
                                     DirOut=DirOut, 
                                     FileOut=FileOut,
                                     CorrectAlsoMonoisotopic=TRUE)
  
  success <- correctionResults$success
  results <- correctionResults$results
  
  testinfo <- paste0('Dataset: ', Project) 
  
  expect_equal(success, ExpectedOutcome, info = testinfo)
  
  #Load reference results and test results from file and compare
  
  for(resultType in names(resultTypes)) {
    
    reference <- loadResults(filepath=file.path(ProjectPath, 'reference', paste0(FileOut, '_', resultTypes[resultType], '.csv')),
                             filetype='csv', dataDescription='reference')
    
    testresults <- results[[resultType]]
    
    testinfoSpecific <- paste0(testinfo, ', result type: ', resultType)
    
    expect_equal(testresults, reference, info = testinfoSpecific)
    
  }
  
  #Cleanup
  
  unlink(DirOut, recursive = TRUE)
     
})

context("Test reading and writing of excel files by IsoCorrectoR")

test_that("Test reading of xls files", {
  
  ExpectedOutcome = 'TRUE'
  
  Project <- 'AA1'
  
  ProjectPath <- file.path(TestDataPath, 'end_to_end_nres', Project)
  
  DirOut = file.path(ProjectPath, 'test_output')
  
  if(!dir.exists(DirOut)) {
    
    dir.create(DirOut)
    
  }
  
  MeasurementFile = file.path(ProjectPath, 'MeasurementFile.xlsx')
  MoleculeFile = file.path(ProjectPath, 'MoleculeFile.xlsx')
  ElementFile = file.path(TestDataPath, 'ElementFile.xlsx')
  
  FileOut = "co_np"
  
  #run IsoCorrection with given parameters
      
  correctionResults <- IsoCorrectionSafeCall(MeasurementFile=MeasurementFile,
                                     ElementFile=ElementFile,
                                     MoleculeFile=MoleculeFile,
                                     DirOut=DirOut, 
                                     FileOut=FileOut,
                                     FileOutFormat="csv",
                                     CorrectAlsoMonoisotopic=TRUE)
  
  success <- correctionResults$success
  results <- correctionResults$results
  
  testinfo <- paste0('Dataset: ', Project) 
  
  expect_equal(success, ExpectedOutcome, info = testinfo)
  
  #Load reference results and test results from file and compare
  
  for(resultType in names(resultTypes)) {
    
    reference <- loadResults(filepath=file.path(ProjectPath, 'reference', paste0(FileOut, '_', resultTypes[resultType], '.csv')),
                             filetype='csv', dataDescription='reference')
    
    testresults <- loadResults(filepath=file.path(ProjectPath, 'test_output', paste0('IsoCorrectoR_', FileOut, '_', resultType, '.csv')),
                               filetype='csv', dataDescription='testresults')
    
    testinfoSpecific <- paste0(testinfo, ', result type: ', resultType)
    
    expect_equal(testresults, reference, info = testinfoSpecific)
    
  }
  
  #Cleanup
  
  unlink(DirOut, recursive = TRUE)
     
})

test_that("Test writing of xls files", {
  
  skip('Skipped')
  
  ExpectedOutcome = 'TRUE'
  
  Project <- 'AA1'
  
  ProjectPath <- file.path(TestDataPath, 'end_to_end_nres', Project)
  
  DirOut = file.path(ProjectPath, 'test_output')
  
  if(!dir.exists(DirOut)) {
    
    dir.create(DirOut)
    
  }
  
  MeasurementFile = file.path(ProjectPath, 'MeasurementFile.csv')
  MoleculeFile = file.path(ProjectPath, 'MoleculeFile.csv')
  ElementFile = file.path(TestDataPath, 'ElementFile.csv')
  
  FileOut = "co_np"
  
  #run IsoCorrection with given parameters
  
  correctionResults <- IsoCorrectionSafeCall(MeasurementFile=MeasurementFile,
                                             ElementFile=ElementFile,
                                             MoleculeFile=MoleculeFile,
                                             DirOut=DirOut, 
                                             FileOut=FileOut,
                                             FileOutFormat="xls",
                                             CorrectAlsoMonoisotopic=TRUE)
  
  success <- correctionResults$success
  results <- correctionResults$results
  
  testinfo <- paste0('Dataset: ', Project) 
  
  expect_equal(success, ExpectedOutcome, info = testinfo)
  
  #Load reference results and test results from file and compare
  
  for(resultType in names(resultTypes)) {
    
    reference <- loadResults(filepath=file.path(ProjectPath, 'reference', paste0(FileOut, '_', resultTypes[resultType], '.csv')),
                             filetype='csv', dataDescription='reference')
    
    testresults <- loadResults(filepath=file.path(ProjectPath, 'test_output', paste0('IsoCorrectoR_', FileOut, '.xls')),
                               filetype='xls', sheet = resultType, dataDescription='testresults')
    
    testinfoSpecific <- paste0(testinfo, ', result type: ', resultType)
    
    expect_equal(testresults, reference, info = testinfoSpecific)
    
  }
  
  #Cleanup
  
  unlink(DirOut, recursive = TRUE)
  
})

context("End to end testing of IsoCorrectoR with csv files")

ElementFile = file.path(TestDataPath, 'ElementFile.csv')

endToEndTest <- function(PathToProjects, Project, ElementFilePath,
                         ResolutionOption=FALSE, ExpectedOutcome='TRUE', resultTypes=NA) {
  
  ProjectPath <- file.path(PathToProjects, Project)
  
  MeasurementFile = file.path(ProjectPath, 'MeasurementFile.csv')
  MoleculeFile = file.path(ProjectPath, 'MoleculeFile.csv')
  
  ImpurityOptions <- c(TRUE, FALSE)
  CoreOptions <- c(TRUE, FALSE)
  names(ImpurityOptions) <- c('p', 'np')
  names(CoreOptions) <- c('co', 'nco')
  
  DirOut = file.path(ProjectPath, 'test_output')
  
  if(!dir.exists(DirOut)) {
    
    dir.create(DirOut)
    
  }
  
  for (ImpurityOption in names(ImpurityOptions)) {
    
    for (CoreOption in names(CoreOptions)) {
      
      FileOut = paste0(CoreOption, '_', ImpurityOption)
      
      #run IsoCorrection with given parameters
      
      correctionResults <- IsoCorrectionSafeCall(MeasurementFile=MeasurementFile,
                                         ElementFile=ElementFilePath,
                                         MoleculeFile=MoleculeFile,
                                         CorrectTracerImpurity=ImpurityOptions[ImpurityOption],
                                         CorrectTracerElementCore=CoreOptions[CoreOption], 
                                         UltraHighRes=ResolutionOption,
                                         DirOut=DirOut, 
                                         FileOut=FileOut,
                                         CorrectAlsoMonoisotopic=TRUE,
                                         Testmode=TRUE)
      
      success <- correctionResults$success
      results <- correctionResults$results
      
      testinfo <- paste0('Dataset: ', Project, ', purity: ', ImpurityOptions[ImpurityOption], ', core: ', CoreOptions[CoreOption]) 
      
      expect_equal(success, ExpectedOutcome, info = testinfo)
      
      #Load reference results and test results from file and compare
      
      for(resultType in names(resultTypes)) {
        
        reference <- loadResults(filepath=file.path(ProjectPath, 'reference', paste0(FileOut, '_', resultTypes[resultType], '.csv')),
                                 filetype='csv', dataDescription='reference')
        
        testresults <- loadResults(filepath=file.path(ProjectPath, 'test_output', paste0('IsoCorrectoR_', FileOut, '_', resultType, '.csv')),
                                   filetype='csv', dataDescription='testresults')
        
        testinfoSpecific <- paste0(testinfo, ', result type: ', resultType)
        
        expect_equal(testresults, reference, info = testinfoSpecific)
        
      }
    }
  }
  
  #Cleanup
  
  unlink(DirOut, recursive = TRUE)
  
}

test_that("End to end tests for normal resolution", {
  
  generalPath <- file.path(TestDataPath, 'end_to_end_nres')

  ProjectFolders <- list.files(generalPath)
  
  for (Project in ProjectFolders) {
    
    endToEndTest(PathToProjects=generalPath, Project=Project, ElementFilePath=ElementFile,
                 ResolutionOption=FALSE, resultTypes = resultTypes)
    
  }
     
})

test_that("End to end tests for high resolution", {
  
  generalPath <- file.path(TestDataPath, 'end_to_end_hres')
  
  ProjectFolders <- list.files(generalPath)
  
  for (Project in ProjectFolders) {
    
    endToEndTest(PathToProjects=generalPath, Project=Project, ElementFilePath=ElementFile,
                             ResolutionOption=TRUE, resultTypes = resultTypes)
      
  }
  
})

test_that("End to end tests for normal resolution with missing values", {
  
  generalPath <- file.path(TestDataPath, 'missing_values_nres')
  
  ProjectFolders <- list.files(generalPath)
  
  for (Project in ProjectFolders) {
    
    suppressWarnings(endToEndTest(PathToProjects=generalPath, Project=Project, ElementFilePath=ElementFile,
                 ResolutionOption=FALSE, ExpectedOutcome = "WARNINGS", resultTypes = resultTypes))
    
  }
  
})

test_that("End to end tests for normal resolution, additional testsets", {
    
  skip('Skipped')
  
  generalPath <- file.path(TestDataPath, 'additional_testsets_nres')
  
  ProjectFolders <- list.files(generalPath)
  
  for (Project in ProjectFolders) {
    
    endToEndTest(PathToProjects=generalPath, Project=Project, ElementFilePath=ElementFile,
                 ResolutionOption=FALSE, resultTypes = resultTypes)
    
  }
  
})

# COMPUTE ISOTOPE COMBINATIONS OF NON-TRACER ELEMENT

# Combinations, probabilites and mass shifts of the non-tracer elements are
# computed, given that not only the tracer element is provided in
# the molcule file. As with the tracer element,
# IsoCombinationsMaster() feeds into IsoCombinations() also for the other elements.

IsoCombinationsNonTracer <- function(MoleculeFragmentData, Fragment, ElementInfo,
                                     NumberTracers, IDTracer, CalculationThreshold) {
  ProbElemList <- list()
  MassShiftElemList <- list()
  NumberIsoCombEff <- list()
  IsotopesElem <- list()

  ElementNonTracer_label <- "NonTracer"
  ElementsNonTracer <- MoleculeFragmentData[[ElementNonTracer_label]]

  if (length(ElementsNonTracer) == 0) {
    ElementsNonTracer <- MoleculeFragmentData[["ZeroTracer"]]
    ElementNonTracer_label <- "ZeroTracer"
  }

  NumberElementsNonTracer <- length(ElementsNonTracer)

  if (NumberElementsNonTracer > 0) {
    for (NonTracer in seq_len(NumberElementsNonTracer)) {

      # Definition of parameters for the non-tracer elements
      x <- NonTracer

      Element <- names(ElementsNonTracer)[x]

      if (NumberTracers > 0) {
        PlacesToAssign <- as.numeric(ElementInfo[[IDTracer]][[2]] * MoleculeFragmentData[["MaxLabel"]])
      } else {
        PlacesToAssign <- 0
      } # NumberTracers > 0

      AvailablePlacesTotal <- as.numeric(MoleculeFragmentData[[ElementNonTracer_label]][Element])

      if (PlacesToAssign <= AvailablePlacesTotal) {
        IsoCluster <- PlacesToAssign
      } else {
        IsoCluster <- AvailablePlacesTotal
      }

      # Calling of IsoCombinations()
      IsoCombinationsResultList <- IsoCombinations(
        Fragment = Fragment, x = x, Element = Element, ElementInfo = ElementInfo,
        AvailablePlacesTotal = AvailablePlacesTotal, IsoCluster = IsoCluster, CalculationThreshold = CalculationThreshold
      )
      NumberIsoComb <- IsoCombinationsResultList[["NumberIsoComb"]]
      tmpProbElem_Array <- vector()
      tmpMassShiftElem_Array <- vector()

      # tmpIsotopesElem <- list()

      for (i in seq_len(NumberIsoComb)) {
        tmpProbElem_Array[i] <- IsoCombinationsResultList[["ProbArray"]][i]
        tmpMassShiftElem_Array[i] <- IsoCombinationsResultList[["MassShiftArray"]][i]
      } # i

      ProbElemList[[x]] <- tmpProbElem_Array
      MassShiftElemList[[x]] <- tmpMassShiftElem_Array
      NumberIsoCombEff[[x]] <- NumberIsoComb
      IsotopesElem[[x]] <- IsoCombinationsResultList[["Isotopes"]]
    } # NonTracer

    names(ProbElemList) <- names(ElementsNonTracer)
    names(MassShiftElemList) <- names(ElementsNonTracer)
    names(NumberIsoCombEff) <- names(ElementsNonTracer)
    names(IsotopesElem) <- names(ElementsNonTracer)
  } # NumberElementsNonTracer > 0

  return(list(
    "ProbElemList" = ProbElemList,
    "MassShiftElemList" = MassShiftElemList,
    "NumberIsoCombEff" = NumberIsoCombEff,
    "IsotopesElem" = IsotopesElem
  ))
}

# Calculate probabilites for net tracer counts for UHR

# Similarly to the function MassShiftProbabilities in the normal resolution 
# mode, this part of the code finds all probabilites of a certain tracer in 
# the probability list NatAbuImpurityProbList that are associated with a total 
# number of tracer isotope TotalLabel.
# It then sums them up to yield CumProbList.

TracerCountProbabilitiesUHR <- function(MoleculeData,NatAbuImpurityList, verbose) {

  NumberTracers <- length(MoleculeData[[1]][["Tracer"]])
  MaxLabel <- MoleculeData[[1]][["MaxLabel"]]
  
  NatAbuImpurityShiftList <- NatAbuImpurityList$NatAbuImpurityShiftList
  NatAbuImpurityProbList <- NatAbuImpurityList$NatAbuImpurityProbList
  
  CumProbList <- list()
  for (TracerNo in seq_len(NumberTracers)) {
    if(verbose){message(date(), " :: [TracerNo] ", TracerNo)}
    
    tmpCumProbList <- list()
    for (IntrinsicLabel in 0:MaxLabel[TracerNo]) {
      if(verbose){message(date(), " :: [IntrinsicLabel] ", IntrinsicLabel)}
      tmpCumProb.vec <- vector()
      for (TotalLabel in 0:MaxLabel[TracerNo]) {
        if(verbose){message(date(), " :: [TotalLabel] ", TotalLabel)}
        
        tmp.idx <- lapply(NatAbuImpurityShiftList[[TracerNo]][[IntrinsicLabel + 1]], function(x) x == TotalLabel)
        
        tmpProb <- unlist(NatAbuImpurityProbList[[TracerNo]][[IntrinsicLabel + 1]])[unlist(tmp.idx)]
        
        tmpCumProb.vec[TotalLabel + 1] <- sum(unlist(lapply(tmpProb, sum)))
        
      }  #TotalLabel
      
      tmpCumProbList[[IntrinsicLabel + 1]] <- tmpCumProb.vec
    }  #IntrinsicLabel
    
    CumProbList[[TracerNo]] <- tmpCumProbList
    
  }  #TracerNo
  
  return(CumProbList)
  
}

#' Algorithm For Natural Isotope Abundance And Tracer Purity Correction Of Data From Stable Isotope Labeling Experiments
#'
#' \code{IsoCorrection} is the main function of the IsoCorrectoR package.
#' It performs the correction of mass spectrometry data from stable isotope
#' labeling experiments with regard to natural abundance and tracer purity.
#' Data from both MS and MS/MS experiments can be corrected
#' (with any tracer isotope: 13C, 15N, 18O...), as well as high resolution data
#' from multiple-tracer experiments (e.g. 13C and 15N used simultaneously).
#'
#' @param MeasurementFile Required. The file that contains the measured data to
#'   be corrected. Only ".xls",".xlsx" and ".csv" file formats are supported.
#'
#' @param ElementFile Required. The file that contains the element information
#'   required for correction. Only ".xls",".xlsx" and ".csv" file formats are
#'   supported.
#'
#' @param MoleculeFile Required. The file that contains the information on the
#'   molecules for which data is to be corrected. Only ".xls",".xlsx" and ".csv"
#'   file formats are supported.
#'
#' @param CorrectTracerImpurity Logical. If TRUE, correction for isotopic
#'   impurity of the tracer substrate is performed.
#'
#' @param CorrectTracerElementCore Logical. If TRUE (recommended!), the tracer
#'   element atoms in the core module (usually the part of the molecule that
#'   does not come from derivatization) are considered when correcting.
#'
#' @param CalculateMeanEnrichment Logical. If TRUE, the mean isotopic enrichment
#'   is calculated for each molecule.
#'
#' @param UltraHighRes Logical. If TRUE, high resolution correction is performed
#'   on the data. Should only be set to TRUE, if you know that you have high
#'   resolution data.
#'
#' @param DirOut Character String. Defines the directory the corrected data and
#'   log-file should be written to. Default directory is set to current working
#'   directoy ('.').
#'
#' @param FileOut Character String. Defines the name of the file that contains
#'   the corrected data. The name of the fill will be
#'   IsoCorrectoR_<FileOut>.<FileFormat>. If the format is set to "csv", the
#'   name will also contain the type of the corrected data in the respective
#'   file.
#'
#' @param FileOutFormat Character String. Defines the format of the files that
#'   contain the corrected data. Can either be "csv" or "xls". If set to "csv",
#'   multiple files will be generated, one for each type of corrected data (eg.
#'   corrected data, fractions, mean enrichment...). If set to "xls", all
#'   correction results are provided in one excel file in different sheets.
#'
#' @param ReturnResultsObject Logical. If TRUE, the correction results are
#'   returned as a list in the current R_session in addition to writing the
#'   results to a file. This is useful if the corrected data has to be further
#'   processed directly in R.
#'
#' @param CorrectAlsoMonoisotopic Logical. If TRUE, monoisotopic correction
#'   results are also provided.
#'
#' @param CalculationThreshold (Advanced Option) Numeric. Defines a threshold to
#'   stop probability calculations at for making correction faster (normal
#'   resolution mode). Should be left at the default value.
#'
#' @param CalculationThreshold_UHR (Advanced Option) Numeric. Defines a
#'   threshold to stop probability calculations at for making correction faster
#'   (high resolution mode). Should be left at the default value.
#'
#' @param verbose Logical. If TRUE, status messages are sent to standard output.
#'
#' @param Testmode Logical. If TRUE, starts a testmode for development purposes.
#'   Not required for users of IsoCorrectoR.
#'
#'
#' @import quadprog utils
#'
#' @importFrom WriteXLS WriteXLS
#' @importFrom stringr str_detect
#' @importFrom utils write.csv
#' @importFrom utils packageVersion
#'
#' @export
#'
#' @examples
#'
#' # Normal resolution data
#'
#'    # 1) get path of example files
#'    path.molecule <- system.file("extdata","normal_resolution","MoleculeFile.csv",
#'       package = "IsoCorrectoR", mustWork = TRUE);
#'    path.element <- system.file("extdata","normal_resolution","ElementFile.csv",
#'       package = "IsoCorrectoR", mustWork = TRUE);
#'    path.measurement <- system.file("extdata","normal_resolution","MeasurementFile.csv",
#'       package = "IsoCorrectoR", mustWork = TRUE);
#'
#'    # 2) run correction algorithm and save results in variable
#'    correctionResults <- IsoCorrection(MeasurementFile=path.measurement,
#'       ElementFile=path.element,
#'      MoleculeFile=path.molecule)
#'
#' # High resolution data
#'
#'    # 1) get path of example files
#'    path.molecule <- system.file("extdata","high_resolution","MoleculeFile.csv",
#'       package = "IsoCorrectoR", mustWork = TRUE);
#'    path.element <- system.file("extdata","high_resolution","ElementFile.csv",
#'       package = "IsoCorrectoR", mustWork = TRUE);
#'    path.measurement <- system.file("extdata","high_resolution","MeasurementFile.csv",
#'       package = "IsoCorrectoR", mustWork = TRUE);
#'
#'    # 2) run correction algorithm and save results in variable
#'    correctionResults <- IsoCorrection(MeasurementFile=path.measurement,
#'       ElementFile=path.element,
#'      MoleculeFile=path.molecule,UltraHighRes=TRUE)
#'
#' @return The function returns a list with 4 elements
#'
#' \describe{
#'  \item{success: }{string that is "TRUE" if the correction was successful, "FALSE" if an error occured and "WARNINGS"
#'  if warnings occured}
#'  \item{results: }{a list containing a dataframe for each type of corrected data (normal, fractions, mean enrichment ...).
#'  Will be NA if ReturnResultsObject is set to FALSE}
#'  \item{log: }{list containing log information on the correction run (parameters, file names and paths, warnings and errors)}
#'  \item{error: }{contains a string with the associated error message if an error occurred, empty otherwise}
#' }
#'
#' @references See Reference 1 \url{Link to IsoCorrectoR-Paper}
#'

IsoCorrection <-
  function(MeasurementFile = NA,
             ElementFile = NA,
             MoleculeFile = NA,
             CorrectTracerImpurity = FALSE,
             CorrectTracerElementCore = TRUE,
             CalculateMeanEnrichment = TRUE,
             UltraHighRes = FALSE,
             DirOut = ".",
             FileOut = "result",
             FileOutFormat = "csv",
             ReturnResultsObject = TRUE,
             CorrectAlsoMonoisotopic = FALSE,
             CalculationThreshold = 10^-8,
             CalculationThreshold_UHR = 8,
             verbose=FALSE,
             Testmode = FALSE) {
    message("\n", rep("#", options()$width))
    message(date(), " :: Starting IsoCorrection ...\n")
    # message(rep("#", options()$width))

    version <- utils::packageVersion("IsoCorrectoR")

    timestamp <- Sys.time()
    timestampFMT <- strftime(timestamp, "%Y-%m-%d_%H%M%S")

    correctionOutput <- list()

    correctionOutput$success <- "FALSE"
    correctionOutput$results <- NA
    correctionOutput$log <- NA
    correctionOutput$error <- NA

    roundDigits <- 8

    # DEFINE LOG ENVIRONMENT

    log.env <- new.env()

    log.env$error <- character()
    log.env$warning <- character()
    log.env$general <- character()

    log.env$param <- list()

    log.env$param$Timestamp <- timestamp
    log.env$param$TimestampFMT <- timestampFMT

    log.env$param$version <- version
    log.env$param$date <- date()
    log.env$param$MeasurementFile <- MeasurementFile
    log.env$param$MoleculeFile <- MoleculeFile
    log.env$param$ElementFile <- ElementFile
    log.env$param$NamePrefix <- "IsoCorrectoR"
    log.env$param$WorkingDirectory <- getwd()
    log.env$param$LogfileName <- paste0(log.env$param$NamePrefix, ".log")
    log.env$param$FileOut <- FileOut
    log.env$param$FileOutFormat <- FileOutFormat

    log.env$param$CorrectTracerImpurity <- CorrectTracerImpurity
    log.env$param$CorrectTracerElementCore <- CorrectTracerElementCore
    log.env$param$UltraHighRes <- UltraHighRes
    log.env$param$CalculateMeanEnrichment <- CalculateMeanEnrichment

    if (!UltraHighRes) {
      log.env$param$threshold <- CalculationThreshold
    }
    else {
      log.env$param$threshold <- CalculationThreshold_UHR
    }

    tryCatch({

      # DEFINE OUTPUT DIRECTORY

      if (DirOut == "." || DirOut == "") {
        log.env$param$OutputDirectory <- log.env$param$WorkingDirectory
      }

      else {
        log.env$param$OutputDirectory <- DirOut
      }

      if (Testmode == FALSE) {
        log.env$param$OutputDirectory <-
          file.path(log.env$param$OutputDirectory, timestampFMT)
      }

      ifelse(
        !dir.exists(log.env$param$OutputDirectory),
        dir.create(log.env$param$OutputDirectory, recursive = TRUE),
        FALSE
      )

      # CHECK FUNCTION PARAMETERS

      checkIsoCorrectionParameters(
        MeasurementFile = MeasurementFile,
        ElementFile = ElementFile,
        MoleculeFile = MoleculeFile,
        CorrectTracerImpurity = CorrectTracerImpurity,
        CorrectTracerElementCore = CorrectTracerElementCore,
        CalculateMeanEnrichment = CalculateMeanEnrichment,
        UltraHighRes = UltraHighRes,
        FileOut = FileOut,
        FileOutFormat = FileOutFormat,
        ReturnResultsObject = ReturnResultsObject,
        CorrectAlsoMonoisotopic = CorrectAlsoMonoisotopic,
        CalculationThreshold = CalculationThreshold,
        CalculationThreshold_UHR = CalculationThreshold_UHR,
        logEnvironment = log.env,
        Testmode = Testmode
      )

      # DEFINE NAMES OF OUTPUT FILES

      NamesCorrectedOutput <- defineOutputFilenames(
        logEnvironment = log.env, FileOut = FileOut,
        FileOutFormat = FileOutFormat,
        CalculateMeanEnrichment = CalculateMeanEnrichment,
        CorrectAlsoMonoisotopic = CorrectAlsoMonoisotopic
      )

      # LOAD INPUT FILES, CHECK THEIR STRUCTURE, EXTRACT INFORMATION AND CHECK 
      # INPUT LOGIC

      InputFileInformation <- ExtractInputFileInformation(
        ElementFile = ElementFile, MoleculeFile = MoleculeFile,
        MeasurementFile = MeasurementFile, UltraHighRes = UltraHighRes,
        CorrectTracerImpurity = CorrectTracerImpurity,
        logEnvironment = log.env,
        verbose = verbose
      )

      MoleculeInfo <- InputFileInformation$MoleculeInfo
      ElementInfo <- InputFileInformation$ElementInfo
      MeasurementInfo <- InputFileInformation$MeasurementInfo

      MoleculesTotal <- length(MoleculeInfo)
      SamplesTotal <- ncol(MeasurementInfo)

      # CALCULATION OF PROBABILITY MATRIX

      # In this section, the probability matrix used for data correction 
      # is produced.

      if (verbose) {
        message(
          date(),
          " :: preparing ",
          MoleculesTotal,
          " molecules for correction ..."
        )
      }

      # if(verbose){message(date(), " :: UltraHighRes==", UltraHighRes)}
      if (verbose) {
        if (UltraHighRes == FALSE) {
          message(date(), " :: running in normal resolution mode.")
        } else {
          message(date(), " :: running in ultra high resolution mode,")
        }
      }


      # NORMAL RESOLUTION

      if (!UltraHighRes) {
        CombinedProbList <- ProbNormalRes(
          MoleculeInfo = MoleculeInfo,
          MoleculesTotal = MoleculesTotal,
          ElementInfo = ElementInfo,
          CorrectTracerElementCore = CorrectTracerElementCore,
          CorrectTracerImpurity = CorrectTracerImpurity,
          CalculationThreshold = CalculationThreshold,
          verbose = verbose
        )

        # HIGH RESOLUTION
      } else if (UltraHighRes) {
        CombinedProbList <- ProbUltraHighRes(
          MoleculeInfo = MoleculeInfo,
          MoleculesTotal = MoleculesTotal,
          ElementInfo = ElementInfo,
          CorrectTracerElementCore = CorrectTracerElementCore,
          CorrectTracerImpurity = CorrectTracerImpurity,
          CalculationThreshold_UHR = CalculationThreshold_UHR,
          verbose = verbose
        )
      } # (!UltraHighRes)

      # DATA CORRECTION USING THE PROBABILITY MATRIX

      CorrectedData <- RawDataCorrectionOnCompleteDataset(
        MeasurementInfo = MeasurementInfo,
        MoleculeInfo = MoleculeInfo,
        CombinedProbList = CombinedProbList,
        MoleculesTotal = MoleculesTotal,
        SamplesTotal = SamplesTotal,
        NamesCorrectedOutput = NamesCorrectedOutput,
        roundDigits = roundDigits,
        logEnvironment = log.env,
        verbose = verbose
      )

      CorrectedDataOutput <- CorrectedData$CorrectedDataOutput
      CorrectionResultList <- CorrectedData$CorrectionResultList

      # CALCULATION OF MEAN ENRICHMENT

      if (CalculateMeanEnrichment) {
        MeanEnrichmentDataOutput <-
          CalcMeanEnrichment(
            MoleculeInfo = MoleculeInfo,
            MoleculesTotal = MoleculesTotal,
            SamplesTotal = SamplesTotal,
            roundDigits = roundDigits,
            CorrectionResultList = CorrectionResultList,
            UltraHighRes = UltraHighRes,
            verbose = verbose
          )

        CorrectedDataOutput$MeanEnrichment <- MeanEnrichmentDataOutput
      }

      # WRITE CORRECTED DATA TO FILE

      writeCorrectionResultsToFile(
        CorrectedDataOutput = CorrectedDataOutput,
        NamesCorrectedOutput = NamesCorrectedOutput,
        logEnvironment = log.env,
        verbose = verbose
      )

      # END OF CORRECTION

      if (ReturnResultsObject) {
        correctionOutput$results <- CorrectedDataOutput
      }
      
      if(length(log.env$warning) == 0) {
      
        correctionOutput$success <- "TRUE"
        
      } else {
        
        correctionOutput$success <- "WARNINGS"
        
      }

      # verbose independent message
      # message("\n\n", rep("#", options()$width))
      message(date(), " :: IsoCorrection successfully completed!")
      message(rep("#", options()$width), "\n")
    },

    # ERROR HANDLING
    error = function(e) {
      commonErrorString <-
        "\n\nTHE CORRECTION PROCEDURE WAS ABORTED BECAUSE AN ERROR HAS OCCURED.
      PLEASE CHECK YOUR LOG-FILE.\n\n"

      log.env$error <- c(log.env$error, e)
      
      # verbose independent message
      
      message(paste0(commonErrorString, e))
        
    }
    )

    tryCatch({
      writeLog(log.env)
    },
    error = function(e) {
      log.env$error <- c(log.env$error, e)
      
      # verbose independent message
      
      message(paste0("THE LOG-FILE COULD NOT BE WRITTEN. ERROR: ", e[[1]]))
      
    }
    )

    correctionOutput$log <- as.list(log.env)

    correctionOutput$error <- log.env$error[1][[1]]

    return(correctionOutput)
  } # IsoCorrection

# Computation of isotope combinations, associated probabilities and mass shifts 
# for an element

#' @importFrom stringr str_c

IsoCombinationsMaster <- function(MoleculeData, MoleculeName, MoleculeNo, Fragment, ElementInfo, CalculationThreshold, CorrectTracerImpurity, CorrectTracerElementCore) {
  
  MoleculeFragmentData <- MoleculeData[[stringr::str_c("Fragment_", Fragment)]]  # get individual fragment data
  
  Transitions <- MoleculeData[["Transitions"]]
  
  NumberTransitions <- nrow(Transitions)
  
  NumberTracers <- length(MoleculeFragmentData[["Tracer"]])
  
  # Isotope combinations for tracer element
  
  IsoCombinationsTracerResult <- IsoCombinationsTracer(MoleculeFragmentData=MoleculeFragmentData,
                                                       Fragment=Fragment,
                                                       ElementInfo = ElementInfo,
                                                       Transitions=Transitions,
                                                       NumberTracers=NumberTracers,
                                                       CorrectTracerImpurity=CorrectTracerImpurity, 
                                                       CorrectTracerElementCore=CorrectTracerElementCore,
                                                       CalculationThreshold=CalculationThreshold)
  
  ProbTracerList <- IsoCombinationsTracerResult$ProbTracerList
  MassShiftTracerList <- IsoCombinationsTracerResult$MassShiftTracerList
  NumberIsoCombTracerEff <- IsoCombinationsTracerResult$NumberIsoCombTracerEff
  IsotopesTracer <- IsoCombinationsTracerResult$IsotopesTracer
  LabelPresent <- IsoCombinationsTracerResult$LabelPresent
  
  # Isotope combinations for non-tracer elements
  
  if (NumberTracers > 0) {
    
    IDTracer <- MoleculeFragmentData[["IDTracer"]] 
    
  } else {
    
    IDTracer <- NA
    
  }
  
  IsoCombinationsNonTracerResult <- IsoCombinationsNonTracer(MoleculeFragmentData=MoleculeFragmentData, 
                                                            Fragment=Fragment,
                                                            ElementInfo=ElementInfo,
                                                            NumberTracers=NumberTracers,
                                                            IDTracer=IDTracer,
                                                            CalculationThreshold=CalculationThreshold)
  
  ProbElemList <- IsoCombinationsNonTracerResult$ProbElemList
  MassShiftElemList <- IsoCombinationsNonTracerResult$MassShiftElemList
  NumberIsoCombEff <- IsoCombinationsNonTracerResult$NumberIsoCombEff
  IsotopesElem <- IsoCombinationsNonTracerResult$IsotopesElem

  # Compute tracer impurity combinations
  
  ProbTracerImpurityList <- list()
  MassShiftTracerImpurityList <- list()
  NumberImpurityCombEff <- list()
  
  if (CorrectTracerImpurity) {
    
    if (NumberTracers > 0) {
      
      IsoCombinationsImpurityResult <- IsoCombinationsTracerImpurity(MoleculeFragmentData=MoleculeFragmentData,
                                                                     ElementInfo=ElementInfo,
                                                                     Transitions=Transitions,
                                                                     NumberTransitions=NumberTransitions,
                                                                     IDTracer=IDTracer,
                                                                     LabelPresent=LabelPresent,
                                                                     CalculationThreshold=CalculationThreshold)
      
      ProbTracerImpurityList <- IsoCombinationsImpurityResult$ProbTracerImpurityList
      MassShiftTracerImpurityList <- IsoCombinationsImpurityResult$MassShiftTracerImpurityList
      NumberImpurityCombEff <- IsoCombinationsImpurityResult$NumberImpurityCombEff
      
    } 
  } else {
    ProbTracerImpurityList <- "No Correction for Tracer Impurity performed"
    MassShiftTracerImpurityList <- "No Correction for Tracer Impurity performed"
    NumberImpurityCombEff <- "No Correction for Tracer Impurity performed"
  } #CorrectTracerImpurity=TRUE
  
  IsoCombinationsResult <- list(ProbTracerList = ProbTracerList, ProbElemList = ProbElemList, ProbTracerImpurityList = ProbTracerImpurityList, MassShiftTracerList = MassShiftTracerList, 
                     MassShiftElemList = MassShiftElemList, MassShiftTracerImpurityList = MassShiftTracerImpurityList, NumberIsoCombEff = unlist(NumberIsoCombEff), 
                     NumberImpurityCombEff = NumberImpurityCombEff, NumberIsoCombTracerEff = unlist(NumberIsoCombTracerEff), IsotopesTracer = IsotopesTracer, IsotopesElem = IsotopesElem, 
                     LabelPresent = LabelPresent)
  
  return(IsoCombinationsResult)
  
}  #IsoCombinationsMaster()

# COMBINATION WITH TRACER IMPURITY STATES

# If correction for tracer impurity is switched on, the arrays TracerElemProbList and TracerElemMassShiftList are additonally combined with the
# probabilites and mass shifts associated with different numbers of 'impure' tracer atoms in the molecule(-fragment). This yields the arrays
# TracerImpurityCombProbList and TracerImpurityCombMassShiftList. They give the probability and mass shift associated with defined elemental isotope
# combinations and a defined number of 'impure' tracer atoms.

ElementCombinationsTracerImpurity <- function(ProbTracerImpurityList,MassShiftTracerImpurityList,NumberImpurityCombEff,
                                              TracerElemProbList,TracerElemMassShiftList,NumberTracerElemComb,
                                              Transitions,NumberTransitions,CalculationThreshold) {
  
  TracerImpurityCombProbList <- list()
  TracerImpurityCombMassShiftList <- list()
  NumberTracerImpurityComb <- vector()
  
  for (TransitionNo in seq_len(NumberTransitions)) {
    
    tmpTracerImpurityCombProb_Array <- vector()
    tmpTracerImpurityCombMassShift_Array <- vector()
    ImpurityComb <- 1
    for (ImpureTracer in 0:NumberImpurityCombEff[[TransitionNo]]) {
      
      for (EleComb in seq_len(NumberTracerElemComb[TransitionNo])) {
        
        TracerImpurityCombProb <- ProbTracerImpurityList[[TransitionNo]][ImpureTracer + 1] * TracerElemProbList[[TransitionNo]][EleComb]
        
        if (TracerImpurityCombProb > CalculationThreshold) {
          
          tmpTracerImpurityCombProb_Array[ImpurityComb] <- TracerImpurityCombProb
          tmpTracerImpurityCombMassShift_Array[ImpurityComb] <- MassShiftTracerImpurityList[[TransitionNo]][ImpureTracer + 1] + TracerElemMassShiftList[[TransitionNo]][EleComb]
          
          NumberTracerImpurityComb[TransitionNo] <- ImpurityComb
          ImpurityComb <- ImpurityComb + 1
        } else {
          EleComb <- NumberTracerElemComb[TransitionNo] + 1
        }  #TracerImpurityCombProb>CalculationThreshold
      }  #EleComb
    }  #ImpureTracer
    TracerImpurityCombProbList[[TransitionNo]] <- tmpTracerImpurityCombProb_Array
    TracerImpurityCombMassShiftList[[TransitionNo]] <- tmpTracerImpurityCombMassShift_Array
  }  #TransitionNo
  
  names(TracerImpurityCombProbList) <- rownames(Transitions)
  names(TracerImpurityCombMassShiftList) <- rownames(Transitions)
  names(NumberTracerImpurityComb) <- rownames(Transitions)
  
  return(list("TracerImpurityCombProbList"=TracerImpurityCombProbList,
              "TracerImpurityCombMassShiftList"=TracerImpurityCombMassShiftList,
              "NumberTracerImpurityComb"=NumberTracerImpurityComb))
  
}

# Calculate probabilites and mass shifts of element combinations

# With the probabilities and mass shifts of the different
# isotope combinations calculated, the next step is to combine
# the isotope combination probabilities and mass shifts of
# the different elements (and the tracer impurity, if checked) using the function
# 'ElementCombinations'. For a given molecule(-fragment) the function combinatorically
# multiplies all isotope combination probabilities of all non-tracer elements,
# the tracer element and the tracer impurity with each other
# In the same way, the isotope combination mass shifts are added to each
# other. In the end, the probability that a molecule(-fragment) contains a combination
# of specific element isotope combinations for its various elements
# (and a certain number of 'impure' tracer atoms) is yielded,
# together with the mass shift associated with
# such a combination. Because combinatorically multiplying all
# isotope combinations is resource intensive, the
# calculation stops at certain probability thresholds defined by the
# parameter CalculationThreshold. In those cases the probability of a combination is so low
# that it does not affect the correction.

ElementCombinations <- function(MoleculeInfo, MoleculeNo, Fragment, IsoCombinationsResult, CalculationThreshold,
    CorrectTracerImpurity) {
    
    MoleculeFragmentData <- MoleculeInfo[[MoleculeNo]][[str_c("Fragment_", Fragment)]]
    ProbTracerList <- IsoCombinationsResult[["ProbTracerList"]]
    MassShiftTracerList <- IsoCombinationsResult[["MassShiftTracerList"]]
    NumberIsoCombTracerEff <- IsoCombinationsResult[["NumberIsoCombTracerEff"]]
    ProbElemList <- IsoCombinationsResult[["ProbElemList"]]
    MassShiftElemList <- IsoCombinationsResult[["MassShiftElemList"]]
    NumberIsoCombEff <- IsoCombinationsResult[["NumberIsoCombEff"]]
    ProbTracerImpurityList <- IsoCombinationsResult[["ProbTracerImpurityList"]]
    MassShiftTracerImpurityList <- IsoCombinationsResult[["MassShiftTracerImpurityList"]]
    NumberImpurityCombEff <- IsoCombinationsResult[["NumberImpurityCombEff"]]
    ElementsNonTracer <- MoleculeFragmentData[["NonTracer"]]
    
    if (length(ElementsNonTracer) == 0) {
        ElementsNonTracer <- MoleculeFragmentData[["ZeroTracer"]]
    }
    
    NumberElementsNonTracer <- length(ElementsNonTracer)
    
    # CALCULATION OF THE NON-TRACER ELEMENT COMBINATIONS
    
    ElementCombinationsNonTracerResult <- ElementCombinationsNonTracer(ProbElemList=ProbElemList,
                                 MassShiftElemList=MassShiftElemList,
                                 NumberIsoCombEff=NumberIsoCombEff,
                                 ElementsNonTracer=ElementsNonTracer,
                                 NumberElementsNonTracer = NumberElementsNonTracer,
                                 CalculationThreshold=CalculationThreshold)
    
    CombElemProbList <- ElementCombinationsNonTracerResult$CombElemProbList
    CombElemMassShiftList <- ElementCombinationsNonTracerResult$CombElemMassShiftList
    NumberEleComb <- ElementCombinationsNonTracerResult$NumberEleComb
    
    # CALCULATION OF THE TRACER ELEMENT - NON-TRACER ELEMENT COMBINATIONS
    
    NumberTracers <- length(MoleculeFragmentData[["Tracer"]])
    Transitions <- MoleculeInfo[[MoleculeNo]][["Transitions"]]
    NumberTransitions <- nrow(Transitions)
    
    ElementCombinationsTracerResult <- ElementCombinationsTracer(ProbTracerList=ProbTracerList,
                                                                       MassShiftTracerList=MassShiftTracerList, 
                                                                       NumberIsoCombTracerEff=NumberIsoCombTracerEff, 
                                                                       CombElemProbList=CombElemProbList, 
                                                                       CombElemMassShiftList=CombElemMassShiftList,
                                                                       NumberEleComb=NumberEleComb,
                                                                       Fragment=Fragment,
                                                                       NumberElementsNonTracer=NumberElementsNonTracer,
                                                                       NumberTracers=NumberTracers, 
                                                                       Transitions=Transitions,
                                                                       NumberTransitions=NumberTransitions, 
                                                                       CalculationThreshold=CalculationThreshold)
    
    TracerElemProbList <- ElementCombinationsTracerResult$TracerElemProbList
    TracerElemMassShiftList <- ElementCombinationsTracerResult$TracerElemMassShiftList
    NumberTracerElemComb <- ElementCombinationsTracerResult$NumberTracerElemComb
    
    # COMBINATION WITH TRACER IMPURITY STATES
    
    if (CorrectTracerImpurity && (NumberTracers > 0)) {
      
      ElementCombinationsTracerImpurityResult <- ElementCombinationsTracerImpurity(ProbTracerImpurityList=ProbTracerImpurityList,
                                                                                   MassShiftTracerImpurityList=MassShiftTracerImpurityList,
                                                                                   NumberImpurityCombEff=NumberImpurityCombEff,
                                                                                   TracerElemProbList=TracerElemProbList,
                                                                                   TracerElemMassShiftList=TracerElemMassShiftList,
                                                                                   NumberTracerElemComb=NumberTracerElemComb,
                                                                                   Transitions=Transitions,
                                                                                   NumberTransitions=NumberTransitions,
                                                                                   CalculationThreshold=CalculationThreshold) 
      
      TracerImpurityCombProbList <- ElementCombinationsTracerImpurityResult$TracerImpurityCombProbList
      TracerImpurityCombMassShiftList <- ElementCombinationsTracerImpurityResult$TracerImpurityCombMassShiftList
      NumberTracerImpurityComb <- ElementCombinationsTracerImpurityResult$NumberTracerImpurityComb

    } else {
      
      TracerImpurityCombProbList <- list()
      TracerImpurityCombMassShiftList <- list()
      NumberTracerImpurityComb <- vector()
      
    } 

    return(list(TracerElemProbList = TracerElemProbList, TracerElemMassShiftList = TracerElemMassShiftList, TracerImpurityCombProbList = TracerImpurityCombProbList, 
        TracerImpurityCombMassShiftList = TracerImpurityCombMassShiftList, NumberTracerImpurityComb = NumberTracerImpurityComb, NumberTracerElemComb = NumberTracerElemComb))
    
}  #ElementCombinations()

# Extraction of molecule parameters

#' @importFrom stringr str_c
#' @importFrom stringr str_detect
#' @importFrom stringr str_replace
#' @importFrom stringr str_extract
#' @importFrom stringr str_extract_all
#' @importFrom magrittr '%>%'
#'  
MoleculeInfoExtraction <- function(MoleculeData, ElementInfo, UltraHighRes, CorrectTracerImpurity, MoleculesFound, logEnvironment, verbose) {
  
  if(verbose){message(date(), " :: processing molecule file ...")}
  
  #Find molecules from measurement file in molecule data
  
  rownames(MoleculeData) <- as.character(MoleculeData[, 1])
  
  MoleculeLocation.vec <- vector()
  MoleculeLocationLabel.vec <- vector()
  
  for (MoleculeFound in MoleculesFound) {
    
    loc <- grep(paste0("^", MoleculeFound, "$"), rownames(MoleculeData))
    
    if (length(loc) == 1) {
      MoleculeLocation.vec <- c(MoleculeLocation.vec, loc)  # simply add current MoleculeName to vector of all valid MoleculeNames
      MoleculeLocationLabel.vec <- c(MoleculeLocationLabel.vec, MoleculeFound)
    } else {
      notification <- stringr::str_c("Molecule '", MoleculeFound, "' from the measurement data was not found in the molecule file.")
      errorHandler(notification, logEnvironment, "error", verbose=verbose)
    }
  }  #MoleculeFound
  
  names(MoleculeLocation.vec) <- MoleculeLocationLabel.vec
  
  MoleculeList <- list()  # big list containing all necessary information on a molecule
  
  # MoleculeLocation.vec contains location indices of found molecules !
  
  #Analyze imported molecule formulas with regex
  
  for (MoleculeNo in seq_len(length(MoleculeLocation.vec))) {
    
    MoleculeLocation <- MoleculeLocation.vec[MoleculeNo]
    
    MoleculeInformation <- MoleculeData[MoleculeLocation,]
    
    if(verbose){message(date(), " :: :: found molecule: #", MoleculeNo, " [", names(MoleculeLocation.vec)[MoleculeNo], "]")}
    # Check if MoleculeFile contains entries in both the product ion/MS ion 
    # and the neutral loss column or only in the product ion/MS ion column.  
    # This yields NumberFragments for a given molecule and determines whether 
    # MS or MS/MS correction will be applied in the following.
    
    FragmentList <- ParseMoleculeInformation(MoleculeInformation=MoleculeInformation,
                                             ElementInfo=ElementInfo,UltraHighRes=UltraHighRes,
                                             verbose=verbose)
    
    MoleculeList[[MoleculeNo]] <- FragmentList
    
  }  # MoleculeNo
  
  names(MoleculeList) <- names(MoleculeLocation.vec)
  
  #Check logic of extracted molecule information
  
  checkMoleculeDataLogic(MoleculeList, UltraHighRes, CorrectTracerImpurity, ElementInfo, logEnvironment, verbose=verbose)
  
  #Provide log information on molecules and tracers
  
  logEnvironment$param$molecules <- names(MoleculeList)
  
  tracerLogInfo <- list()
  tracerNames <- character()
  
  #Get tracerNames, a vector of all tracer elements used
  
  for (molecule in names(MoleculeList)) {
    
    for (fragmentNo in seq_len(sum(stringr::str_detect(names(MoleculeList[[molecule]]), "Fragment")))) {
      
      fragment <- paste0("Fragment_", fragmentNo)
      
      tracers <- MoleculeList[[molecule]][[fragment]]$Tracer
      
      if (length(tracers) > 0) {
        
        tracerNames <- c(tracerNames, names(tracers))
        
      }
      
    }
    
  }
  
  tracerNames <- unique(tracerNames)
  
  for (tracer in tracerNames) {
    
    if (CorrectTracerImpurity == TRUE) {
      
      tracerLogInfo[[tracer]] <- ElementInfo[[tracer]]$Tracer.purity
      
    } else {
      
      tracerLogInfo[[tracer]] <- NA
      
    }
    
  }
  
  logEnvironment$param$tracers <- tracerLogInfo
  
  if(verbose){message(date(), " :: processing molecule file [OK]\n")}
  
  return(MoleculeInfo = MoleculeList)
}  # function()

#' IsoCorrectoR example data
#' 
#' Example data for the IsoCorrectoR package
#' 
#' @format A list with four elements, namely
#' \describe{
#' \item{tool_features: }{overview of correction features as provided by various tools IsoCorrectoR is compared with.}
#' \item{element_file: }{example data for an Element file. This information is independent of the resolution type (normal or high)}
#' \item{normal_resolution: }{list with two elements, containing examples for normal resolution data: molecule information and measurement data}
#' \item{high_resolution: }{list with two elements, containing examples for high resolution data: molecule information and measurement data}
#' }
#' 
#' 
#' @usage 
#' data(IsoCorrectoR)
#' 
#' @author Paul Heinrich, Christian Kohler
"IsoCorrectoR"

# Calculate probabilites for molecule isotopic compositions regarding the tracer elements

IsoCombinationsUHR <- function(MoleculeData,ElementInfo,CorrectTracerElementCore,CorrectTracerImpurity, verbose) {

  NumberTracers <- length(MoleculeData[[1]][["Tracer"]])
  nTracerMax <- MoleculeData[[1]][["nTracerMax"]]
  IDTracer <- MoleculeData[[1]][["IDTracer"]]
  MaxLabel <- MoleculeData[[1]][["MaxLabel"]]
  NatAbuTracer <- MoleculeData[[1]][["NatAbuTracer"]]
  NatAbuBase <- MoleculeData[[1]][["NatAbuBase"]]
  AvailableTracerPlaces <- vector()
  MaxTracerPlaces <- vector()
  variableTracerMaximum <- list()
  
  Transitions <- MoleculeData[["Transitions"]]
  NumberTransitions <- nrow(Transitions)
  
  NatAbuProbList <- list()
  NatAbuShiftList <- list()
  for (TracerNo in seq_len(NumberTracers)) {
    
    if (CorrectTracerElementCore) {
      AvailableTracerPlaces[TracerNo] <- nTracerMax[TracerNo]
      MaxTracerPlaces[TracerNo] <- MaxLabel[TracerNo]
    } else if (!CorrectTracerElementCore) {
      AvailableTracerPlaces[TracerNo] <- nTracerMax[TracerNo] - MaxLabel[TracerNo]
      if (MaxLabel[TracerNo] <= AvailableTracerPlaces[TracerNo]) {
        MaxTracerPlaces[TracerNo] <- MaxLabel[TracerNo]
      } else {
        MaxTracerPlaces[TracerNo] <- AvailableTracerPlaces[TracerNo]
      }
    } else {
      stop(date(), " :: [STOP] invalid `CorrectTracerElementCore` value: ", CorrectTracerElementCore)
    }
    tmpNatAbuProbList <- list()
    tmpNatAbuShiftList <- list()
    
    tmp_variableTracerMaximum.vec <- vector(mode = "numeric", length = MaxLabel[TracerNo] + 1)
    
    for (IntrinsicLabel in 0:MaxLabel[TracerNo]) {
      
      if (CorrectTracerElementCore) {
        x <- IntrinsicLabel
        
      } else if (!CorrectTracerElementCore) {
        x <- 0
      }
      NatAbuProb <- vector()
      NatAbuShift <- vector()
      
      if (!CorrectTracerImpurity) {
        tmp_variableTracerMaximum <- MaxTracerPlaces[TracerNo] - x
      } else if (CorrectTracerImpurity && (MaxTracerPlaces[TracerNo] <= (AvailableTracerPlaces[TracerNo] - x))) {
        tmp_variableTracerMaximum = MaxTracerPlaces[TracerNo]
      } else {
        tmp_variableTracerMaximum = AvailableTracerPlaces[TracerNo] - x
      }
      
      for (NatAbuLabel in 0:tmp_variableTracerMaximum) {
        
        NatAbuProb[NatAbuLabel + 1] <- choose(AvailableTracerPlaces[TracerNo] - x, NatAbuLabel) * NatAbuTracer[TracerNo]^NatAbuLabel * NatAbuBase[TracerNo]^(AvailableTracerPlaces[TracerNo] - 
                                                                                                                                                               NatAbuLabel - x)
        
        NatAbuShift[NatAbuLabel + 1] <- IntrinsicLabel + NatAbuLabel
      }  #NatAbuLabel
      
      tmpNatAbuProbList[[IntrinsicLabel + 1]] <- NatAbuProb
      tmpNatAbuShiftList[[IntrinsicLabel + 1]] <- NatAbuShift
      
      tmp_variableTracerMaximum.vec[x + 1] <- tmp_variableTracerMaximum
      
    }  #IntrinsicLabel
    
    variableTracerMaximum[[TracerNo]] <- tmp_variableTracerMaximum.vec
    
    NatAbuProbList[[TracerNo]] <- tmpNatAbuProbList
    NatAbuShiftList[[TracerNo]] <- tmpNatAbuShiftList
    
  }  #TracerNo
  
  # As in the case of the normal resolution approach, a correction for tracer 
  # impurity can be performed, in this case for all tracers present in the
  # molecule. The amount of impure tracer that may be found depends on the 
  # amount of intrinsic label of this tracer in the molecule.
  
  if (CorrectTracerImpurity) {
    ImpureTracerProbList <- list()
    for (TracerNo in seq_len(NumberTracers)) {
      tmpImpureTracerProbList <- list()
      for (IntrinsicLabel in 0:MaxLabel[TracerNo]) {
        ImpureTracerProb <- vector()
        for (ImpureTracer in 0:IntrinsicLabel) {
          ImpureTracerProb[ImpureTracer + 1] <- choose(IntrinsicLabel, ImpureTracer) * (1 - ElementInfo[[IDTracer[TracerNo]]][[3]])^ImpureTracer * 
            ElementInfo[[IDTracer[TracerNo]]][[3]]^(IntrinsicLabel - ImpureTracer)
        }  #ImpureTracer
        tmpImpureTracerProbList[[IntrinsicLabel + 1]] <- ImpureTracerProb
      }  #IntrinsicLabel
      ImpureTracerProbList[[TracerNo]] <- tmpImpureTracerProbList
    }  #TracerNo
    
    # In the following section, natural abundance and tracer impurity 
    # probabilities are combined multiplicatively. 
    # The total shift in tracer number associated
    # with natural abundance and tracer impurity (negative shift) is derived by 
    # adding/substracting all individual shifts.
    
    NatAbuImpurityProbList <- list()
    NatAbuImpurityShiftList <- list()
    for (TracerNo in seq_len(NumberTracers)) {
      
      tmpIntrinsicLabelList <- list()
      tmpIntrinsicLabelShiftList <- list()
      tmp_variableTracerMaximum.vec <- variableTracerMaximum[[TracerNo]]
      
      for (IntrinsicLabel in 0:MaxLabel[TracerNo]) {
        if(verbose){message(date(), " :: IntrinsicLabel:", IntrinsicLabel)}
        if (CorrectTracerElementCore) {
          x <- IntrinsicLabel
        } else if (!CorrectTracerElementCore) {
          x <- 0
        }
        tmpNatAbuLabelList <- list()
        tmpNatAbuLabelShiftList <- list()
        
        tmp_variableTracerMaximum <- tmp_variableTracerMaximum.vec[x + 1]
        
        for (NatAbuLabel in 0:tmp_variableTracerMaximum) {
          if(verbose){message(date(), " :: NatAbuLabel:", NatAbuLabel)}
          NatAbuImpurityProb <- vector()
          NatAbuImpurityShift <- vector()
          for (ImpureTracer in 0:IntrinsicLabel) {
            if(verbose){message(date(), " :: ImpureTracer:", ImpureTracer)}
            NatAbuImpurityProb[ImpureTracer + 1] <- NatAbuProbList[[TracerNo]][[IntrinsicLabel + 1]][NatAbuLabel + 1] * ImpureTracerProbList[[TracerNo]][[IntrinsicLabel + 
                                                                                                                                                            1]][ImpureTracer + 1]
            NatAbuImpurityShift[ImpureTracer + 1] <- IntrinsicLabel + NatAbuLabel - ImpureTracer
          }
          tmpNatAbuLabelList[[NatAbuLabel + 1]] <- NatAbuImpurityProb
          tmpNatAbuLabelShiftList[[NatAbuLabel + 1]] <- NatAbuImpurityShift
        }  #NatAbuLabel
        tmpIntrinsicLabelList[[IntrinsicLabel + 1]] <- tmpNatAbuLabelList
        tmpIntrinsicLabelShiftList[[IntrinsicLabel + 1]] <- tmpNatAbuLabelShiftList
      }  #IntrinsicLabel
      
      NatAbuImpurityProbList[[TracerNo]] <- tmpIntrinsicLabelList
      NatAbuImpurityShiftList[[TracerNo]] <- tmpIntrinsicLabelShiftList
      
    }  #TracerNo
  } else {
    NatAbuImpurityProbList <- NatAbuProbList
    NatAbuImpurityShiftList <- NatAbuShiftList
  }  #CorrectTracerImpurity==TRUE
  
  return(list("NatAbuImpurityProbList"=NatAbuImpurityProbList,"NatAbuImpurityShiftList"=NatAbuImpurityShiftList))
  
}

CalculateTransitionsNR <- function(MoleculeInfo, ElementInfo, verbose) {
  
  if(verbose){message(date(), " :: calculating transitions ...")}
  MoleculesTotal <- length(MoleculeInfo)
  MoleculesName <- names(MoleculeInfo)
  
  # For all molecules, determine the maximum amount of tracer isotopes expected
  # in product ion or neutral loss.
  for (MoleculeNo in seq_len(MoleculesTotal)) {
    
    ### assemble required data
    MoleculeData <- MoleculeInfo[[MoleculeNo]]  # entire 'MoleculeInfo' data for current molecule
    
    NumberFragments <- sum(stringr::str_detect(names(MoleculeData), "Fragment"))  # %>%sum
    NumberTracers <- unlist(lapply(MoleculeData, function(x) length(x[["Tracer"]])))
    MaxLabel <- unlist(lapply(MoleculeData, function(x) x[["MaxLabel"]]))
    IDTracer <- unlist(lapply(MoleculeData, function(x) x[["IDTracer"]]))
    
    if (NumberFragments == 2) {
      
      #if(verbose){message(date(), " :: [", MoleculesName[MoleculeNo], "] // NumberFragments==", NumberFragments)}
      if(verbose){message(date(), " :: :: molecule [", MoleculesName[MoleculeNo], "] with ",NumberFragments," fragments")}
      if (NumberTracers[[1]] > 0) {
        MaxLabelProduct <- MaxLabel[[1]]
      } else {
        MaxLabelProduct <- 0
      }
      
      if (NumberTracers[[2]] > 0) {
        MaxLabelNeutralLoss <- MaxLabel[[2]]
      } else {
        MaxLabelNeutralLoss <- 0
      }
      if (NumberTracers[[1]] > 0) {
        Tracer <- IDTracer[[1]]
      } else if (NumberTracers[[2]] > 0) {
        Tracer <- IDTracer[[2]]
      } else {
        stop(date(), " :: [CalculateTransitions()] No tracer element specified for molecule ", MoleculesName[MoleculeNo], "!")
      }
      
      MaxLabelPrecursor <- MaxLabelProduct + MaxLabelNeutralLoss
      
      # CALCULATION OF MS/MS TRANSITIONS
      
      TransitionNo <- 0
      LabelNeutralLoss <- 0
      ProductIon_SiteSaturation <- 0
      SiteSaturationCorrection <- 0
      
      # This part of the function computes how label in the precursor ion can be distributed among product ion and neutral loss, within the constraints given by
      # MaxLabelProduct and MaxLabelNeutralLoss. This yields the expected MS/MS transitions.
      
      tmpTransitionsExpectedList <- list()
      tmpTransitionsExpectedListNames.vec <- vector()
      
      for (LabelPrecursor in 0:MaxLabelPrecursor) {
        # Until the label in the precursor has exceeded the amount of MaxLabelNeutralLoss, all label is considered to be in the neutral loss at this stage.
        
        if (LabelNeutralLoss < MaxLabelNeutralLoss) {
          LabelNeutralLoss <- LabelPrecursor
        }
        
        # ProductIon_SiteSaturation is a means to determine by how much the label in the precursor exceeds the maximum capacity of the product ion given in
        # MaxLabelProduct.
        
        ProductIon_SiteSaturation <- (MaxLabelProduct - LabelPrecursor) * (-1)
        if (ProductIon_SiteSaturation > 0) {
          SiteSaturationCorrection <- ProductIon_SiteSaturation
        }
        
        # At this stage, the MS/MS transitions are calculated and written into the molecule information list.  The vector named Precursor contains the mass shift
        # (in relation to M+0) associated with the precursor (Number of label in the precursor multiplied with the tracer isotope mass shift given in
        # ElementList).  The vector named NeutralLoss contains the mass shift of the neutral loss. This is determined in a loop which starts at the maximum
        # possible amount of current LabelPrecursor from the parent loop in the neutral loss. It then decreases the amount of LabelNeutralLoss by 'shifting' label
        # from the neutral loss to the product ion (NLLabel_in_ProductIon). Consequently, the product ion mass shift in the vector named ProductIon is given as
        # the difference between precursor mass shift and neutral loss mass shift of the current loop iteration.  This way all combinations of product ion and
        # neutral loss labelling for a given precursor labelling state are derived. The shifting just described has limits given by MaxLabelProduct.  This is
        # where SiteSaturationCorrection becomes relevant.  If the amount of label in the precursor exceeds the maximum amount of labelling possible in the
        # product ion, it will reduce the number of iterations of the shifting loop by just this difference. In addition, the loop generates specific name tags
        # for each transition of a given molecule by combining the molecule name from the molecules information file with '_T' for transition, followed by the
        # precuror label x and product ion label y as x.y (Name_x.y).
        
        # TransitionNo<-0
        for (NLLabel_in_ProductIon in 0:(LabelNeutralLoss - SiteSaturationCorrection)) {
          
          TransitionNo <- TransitionNo + 1
          
          tmpTrans1 <- LabelPrecursor * ElementInfo[[Tracer]][[2]]
          tmpTrans2 <- (LabelNeutralLoss - NLLabel_in_ProductIon) * ElementInfo[[Tracer]][[2]]
          tmpTrans3 <- tmpTrans1 - tmpTrans2
          
          tmpTransitionsExpectedList[[TransitionNo]] <- list(ProductIon = tmpTrans3, NeutralLoss = tmpTrans2, Precursor = tmpTrans1)
          
          tmpTransitionsExpectedListNames.vec <- c(tmpTransitionsExpectedListNames.vec, str_c(MoleculesName[MoleculeNo], "_", tmpTrans1, ".", tmpTrans3))
          
        }  # NLLabel
        
      }  # LabelPrecursor
      
      names(tmpTransitionsExpectedList) <- tmpTransitionsExpectedListNames.vec
      
      TransitionsExpected.df <- data.frame(ProductIon = as.numeric(unlist(lapply(tmpTransitionsExpectedList, function(x) x["ProductIon"]))), NeutralLoss = as.numeric(unlist(lapply(tmpTransitionsExpectedList, 
                                                                                                                                                                                    function(x) x["NeutralLoss"]))), Precursor = as.numeric(unlist(lapply(tmpTransitionsExpectedList, function(x) x["Precursor"]))))
      
      rownames(TransitionsExpected.df) <- tmpTransitionsExpectedListNames.vec
      
      TransitionsExpected <- TransitionsExpected.df
      
      # In this section the expected MS (not MS/MS) measurements and their name tags are generated for each molecule An MS name tag has the structure 'Name_x'
      
    } else if (NumberFragments == 1) 
    {
      
      if (NumberTracers[[1]] > 0) {
        MaxLabelProduct <- MaxLabel[[1]]
        Tracer <- IDTracer[[1]]
      } else {
        MaxLabelProduct <- 0
        warning(date(), " [WARNING] No tracer element specified for Molecule #", MoleculeNo, ".")
      }
      
      TransitionNo <- 0
      tmpTransitionsExpectedList <- list()
      tmpTransitionsExpectedListNames.vec <- vector()
      for (LabelIon in 0:MaxLabelProduct) {
        TransitionNo <- TransitionNo + 1
        tmpTrans1 <- LabelIon * ElementInfo[[Tracer]][[2]]
        
        tmpTransitionsExpectedList[[TransitionNo]] <- list(ProductIon = tmpTrans1, Precursor = tmpTrans1)
        tmpTransitionsExpectedListNames.vec <- c(tmpTransitionsExpectedListNames.vec, str_c(MoleculesName[MoleculeNo], "_", tmpTrans1))
        
      }  #LabelIon
      names(tmpTransitionsExpectedList) <- tmpTransitionsExpectedListNames.vec
      
      TransitionsExpected.df <- data.frame(ProductIon = as.numeric(unlist(lapply(tmpTransitionsExpectedList, function(x) x["ProductIon"]))), Precursor = as.numeric(unlist(lapply(tmpTransitionsExpectedList, 
                                                                                                                                                                                  function(x) x["Precursor"]))))
      
      
      rownames(TransitionsExpected.df) <- tmpTransitionsExpectedListNames.vec
      
      TransitionsExpected <- TransitionsExpected.df
      
    }  # NumberFragments==1
    MoleculeInfo[[MoleculeNo]][["TransitionsExpected"]] <- TransitionsExpected
  }  #  MoleculeNo
  
  if(verbose){message(date(), " :: calculating transitions [OK]\n")}
  return(MoleculeInfo)
  
}

# Calculates the sum of probabilities with a common mass shift

# The function 'MassShiftProbabilities' yields the
# probability that a given transition of a
# molecule(-fragment) (through naturally occuring isotopes)
# produces a certain mass shift in relation to the same molecule(-fragment)
# containing no isotopes of higher mass. This probability is gained
# by summing all entries of a molecule(-fragment) of a given
# transition in TracerElemProbList that have the same mass shift in
# TracerElemMassShiftList. If correction for tracer impurity is turned
# on and if the molecule(-fragment) in question can contain tracer,
# 'MassShiftProbabilities' uses TracerImpurityCombProbList
# and TracerImpurityCombMassShiftList instead.

#' @importFrom stringr str_c
#' 
MassShiftProbabilities <- function(MoleculeInfo, MoleculeNo, Fragment, MaxMassShift, ElementCombinations, CorrectTracerImpurity) {
  
  MoleculeData <- MoleculeInfo[[MoleculeNo]][[stringr::str_c("Fragment_", Fragment)]]
  
  Transitions <- MoleculeInfo[[MoleculeNo]][["Transitions"]]
  NumberTransitions <- nrow(Transitions)
  NumberTracers <- length(MoleculeData[["Tracer"]])
  
  if (CorrectTracerImpurity) {
    if (NumberTracers > 0) {
      
      ProbabilityList <- ElementCombinations[["TracerImpurityCombProbList"]]
      MassShiftList <- ElementCombinations[["TracerImpurityCombMassShiftList"]]
      
    } else {
      ProbabilityList <- ElementCombinations[["TracerElemProbList"]]
      MassShiftList <- ElementCombinations[["TracerElemMassShiftList"]]
      
    }
    
  } else {
    
    ProbabilityList <- ElementCombinations[["TracerElemProbList"]]
    MassShiftList <- ElementCombinations[["TracerElemMassShiftList"]]
    
  }
  
  # This loop goes through all mass shifts until it reaches the maximum possible
  # mass shift of the current molecule(-fragment).  For each labelling state,
  # it finds indexes in MassShiftList associated with the given mass shift 
  # and uses them to find the corresponding probabilites in ProbabilityList.
  # Those values are eventually summed up to yield the probability that a 
  # certain labelling state produces a certain mass shift in relation to the
  # unlabelled species due to natural abundance and possibly tracer impurity.
  
  CumProbList <- list()
  
  for (MassShift in 0:MaxMassShift) {
    
    tmpCumProb.vec <- vector()
    for (TransitionNo in seq_len(NumberTransitions)) {
      
      tmpIdx <- which(MassShiftList[[TransitionNo]] == MassShift)
      
      tmpCumProb.vec[TransitionNo] <- sum(ProbabilityList[[TransitionNo]][tmpIdx])
      
    }
    CumProbList[[MassShift + 1]] <- tmpCumProb.vec
  }  #MassShift
  
  names(CumProbList) <- stringr::str_c("MassShift_", 0:MaxMassShift)
  
  return(CumProbList)
}

# CALCULATION OF PROBABILITY MATRIX
#
# In this section, the probability matrix used for data correction is produced.
# To this end, for each molecule to be corrected, molecule/fragment
# information from the molecule information input file is used
# to calculate the probability that a certain transition or full MS ion x
# is derived from another transition/full MS ion y due to
# natural stable isotope abundance/tracer purity.

ProbNormalRes <- function(MoleculeInfo,MoleculesTotal,ElementInfo,CorrectTracerElementCore,
                          CorrectTracerImpurity,CalculationThreshold, verbose) {

  if(verbose){message(date(), " :: calculating probability matrix with parameters: ")
  message(date(), " :: :: CorrectTracerElementCore: ", CorrectTracerElementCore)
  message(date(), " :: :: CorrectTracerImpurity: ", CorrectTracerImpurity)
  message(date(), " :: :: CalculationThreshold: ", CalculationThreshold)}
  
  MaxMassShiftList <- list()
  icmResultList <- list()
  ecResultList <- list()
  msprobResultList <- list()
  mstResultList <- list()
  
  CombinedProbList <- list()

  for (MoleculeNo in seq_len(MoleculesTotal)) {
    
    MoleculeData <- MoleculeInfo[[MoleculeNo]]
    MoleculeName <- names(MoleculeInfo[MoleculeNo])
    
    NumberFragments <-
      sum(stringr::str_detect(names(MoleculeData), "Fragment"))
    Transitions <- MoleculeData[["Transitions"]]
    
    tmpIcmResultList <- list()
    tmpEcResultList <- list()
    tmpMsprobResultList <- list()
    tmpMstResultList <- list()
    tmpMaxMassShiftList <- list()
    
    # The following functions are run for both fragments of a molecule
    # in the case of MS/MS measurements or just for one molecule(-fragment) for
    # full-MS measurements.
    
    for (Fragment in seq_len(NumberFragments)) {
      # For each molecule, the expected transitions are checked and MaxMassShift
      # gives the mass shifts of the labelled molecule(-fragment(s)) with the highest mass.
      
      MaxMassShift <- apply(Transitions, 2, max)[Fragment]
      tmpMaxMassShiftList[[Fragment]] <- MaxMassShift
      
      # CALCULATE ISOTOPE COMBINATION PROBABILITIES AND MASS SHIFTS
      
      IsoCombinationsResult <-
        IsoCombinationsMaster(
          MoleculeData = MoleculeData,
          MoleculeName = MoleculeName,
          MoleculeNo = MoleculeNo,
          Fragment = Fragment,
          ElementInfo = ElementInfo,
          CorrectTracerElementCore = CorrectTracerElementCore,
          CorrectTracerImpurity = CorrectTracerImpurity,
          CalculationThreshold = CalculationThreshold
        )
      
      tmpIcmResultList[[Fragment]] <- IsoCombinationsResult
      
      # CALCULATE ELEMENT COMBINATION PROBABILITIES AND MASS SHIFTS
      
      ElementCombinationsResult <-
        ElementCombinations(
          MoleculeInfo = MoleculeInfo,
          MoleculeNo = MoleculeNo,
          Fragment = Fragment,
          IsoCombinationsResult = IsoCombinationsResult,
          CalculationThreshold = CalculationThreshold,
          CorrectTracerImpurity = CorrectTracerImpurity
        )
      
      tmpEcResultList[[Fragment]] <- ElementCombinationsResult
      
      #CALCULATE CUMULATIVE PROBABILITIES OF MASS SHIFTS
      
      CumProbList <-
        MassShiftProbabilities(
          MoleculeInfo = MoleculeInfo,
          MoleculeNo = MoleculeNo,
          Fragment = Fragment,
          ElementCombinations = ElementCombinationsResult,
          MaxMassShift = MaxMassShift,
          CorrectTracerImpurity = CorrectTracerImpurity
        )
      
      tmpMsprobResultList[[Fragment]] <- CumProbList
      
      #GET MASS SHIFTS OF MEASUREMENTS/TRANSITIONS
      
      mstResult <-
        MassShiftTransitions(
          MoleculeInfo = MoleculeInfo,
          MoleculeNo = MoleculeNo,
          Fragment = Fragment
        )
      
      tmpMstResultList[[Fragment]] <- mstResult
      
    }#Fragment
    names(tmpIcmResultList) <-
      paste0("Fragment_", seq(seq_len(NumberFragments)))
    names(tmpEcResultList) <-
      paste0("Fragment_", seq(seq_len(NumberFragments)))
    names(tmpMsprobResultList) <-
      paste0("Fragment_", seq(seq_len(NumberFragments)))
    names(tmpMstResultList) <-
      paste0("Fragment_", seq(seq_len(NumberFragments)))
    names(tmpMaxMassShiftList) <-
      paste0("Fragment_", seq(seq_len(NumberFragments)))
    
    # CALCULATE FINAL PROBABILTIY MATRIX CombFragmentProb
    
    CombFragmentProb <-
      FragmentsCombinedProbability(
        MoleculeInfo = MoleculeInfo,
        MoleculeNo = MoleculeNo,
        CumProbList = tmpMsprobResultList,
        FragmentMassShift =
          tmpMstResultList
      )
    
    CombinedProbList[[MoleculeNo]] <-
      CombFragmentProb
    
    MaxMassShiftList[[MoleculeNo]] <- tmpMaxMassShiftList
    icmResultList[[MoleculeNo]] <- tmpIcmResultList
    ecResultList[[MoleculeNo]] <- tmpEcResultList
    msprobResultList[[MoleculeNo]] <- tmpMsprobResultList
    mstResultList[[MoleculeNo]] <- tmpMstResultList
    
  }#MoleculeNo  
  
  names(MaxMassShiftList) <- names(MoleculeInfo)
  names(icmResultList) <- names(MoleculeInfo)
  names(ecResultList) <- names(MoleculeInfo)
  names(msprobResultList) <- names(MoleculeInfo) # CumProbList
  names(mstResultList) <- names(MoleculeInfo) # FragmentMassShift
  names(CombinedProbList) <-
    names(MoleculeInfo) # CombFragmentProb
  
  if(verbose){message(date(), " :: calculating probability matrix [OK]")}
  
  return(CombinedProbList)
  
}

# Calculation of probability and mass shift associated with an isotope combination

IsoCombinationsProbMassShift <- function(Element, NumberIso, ElementInfo, CombinationsArray, IsoCombPre, AvailablePlacesTotal) {
  
  ProbFactorialTermLog <- lfactorial(AvailablePlacesTotal)
  ProbExpTerm <- 1
  MassShiftIsoComb <- 0
  ElementTable <- data.frame(ElementInfo[[Element]][[1]])
  
  for (IsotopeNo in seq_len(NumberIso)) {
    
    IsotopeCount <- as.numeric(CombinationsArray[IsoCombPre, IsotopeNo])
    IsotopeAbundance <- ElementTable[IsotopeNo, 1]
    IsotopeMassShift <- ElementTable[IsotopeNo, 2]
    
    # Probability factorial term (log factorial as factorial can lead to numbers that are too big)
    ProbFactorialTermLog <- ProbFactorialTermLog - lfactorial(IsotopeCount)
    
    # Probability exponential term
    ProbExpTerm <- ProbExpTerm * (IsotopeAbundance^IsotopeCount)
    
    # mass shift
    MassShiftIsoComb <- MassShiftIsoComb + IsotopeMassShift * IsotopeCount
  }  #IsotopeNo
  
  ProbIsoComb <- exp(ProbFactorialTermLog) * ProbExpTerm
  
  return(list(ProbIsoComb = ProbIsoComb, MassShiftIsoComb = MassShiftIsoComb))
  
}

# DATA CORRECTION USING THE PROBABILITY MATRIX
#
# In this function, for each sample and molecule from the the input raw data the
# measured data will be corrected using the probability matrix generated before.

RawDataCorrectionOnCompleteDataset <- function(MeasurementInfo, MoleculeInfo, CombinedProbList, MoleculesTotal, SamplesTotal, NamesCorrectedOutput, roundDigits, logEnvironment, verbose) {
  if (verbose) {
    message(
      date(),
      " :: preparing ",
      MoleculesTotal,
      " molecules for correction [OK]"
    )

    message("\n", rep("#", options()$width))
    message(
      date(),
      " :: Starting data correction for <",
      SamplesTotal,
      " samples> ..."
    )
    message(rep("#", options()$width), "\n")
  }

  CorrectionResultList <- list()

  for (SampleNo in seq_len(SamplesTotal)) {
    if (verbose) {
      message(
        date(),
        " :: processing Sample #",
        SampleNo,
        " [",
        colnames(MeasurementInfo)[SampleNo],
        "] "
      )
    }

    tmpCorrectionResult <- list()

    for (MoleculeNo in seq_len(MoleculesTotal)) {
      ProbMatrix <- CombinedProbList[[MoleculeNo]]

      MoleculeData <- MoleculeInfo[[MoleculeNo]]
      MoleculeName <- names(MoleculeInfo[MoleculeNo])

      TransitionLocation <- unlist(MoleculeData[["TransitionLocation"]])

      # For the molecule defined by MoleculeNo, get the vector of uncorrected values. The location of the value corresponding to a given transition is given by
      # TransitionLocation and based on finding measurement name tags in the input data (RawDataExtraction).

      UncorrectedData <- MeasurementInfo[TransitionLocation, SampleNo]

      # The function 'RawDataCorrection' performs the actual
      # data correction.
      CorrectionResult <- RawDataCorrection(
        UncorrectedData = UncorrectedData,
        MoleculeData = MoleculeData,
        MoleculeName = MoleculeName,
        ProbMatrix = ProbMatrix,
        MoleculeNo = MoleculeNo,
        SampleNo = SampleNo,
        SampleName = colnames(MeasurementInfo)[SampleNo],
        roundDigits = roundDigits,
        logEnvironment = logEnvironment, verbose = verbose
      )

      tmpCorrectionResult[[MoleculeNo]] <- CorrectionResult
    } # MoleculeNo

    names(tmpCorrectionResult) <- names(MoleculeInfo)

    CorrectionResultList[[SampleNo]] <- tmpCorrectionResult
  } # SampleNo

  names(CorrectionResultList) <- colnames(MeasurementInfo)

  x <- convert2matrix(input = CorrectionResultList)

  CorrectedDataOutput <- list()

  CorrectedDataOutput$RawData <- as.data.frame(MeasurementInfo)

  for (name in NamesCorrectedOutput$CorrectedDataNamesBase) {
    CorrectedDataOutput[[name]] <- as.data.frame(x[[name]])
  }

  return(list("CorrectedDataOutput" = CorrectedDataOutput, "CorrectionResultList" = CorrectionResultList))
}

# Generation of a mass shift vector of labelling states

# The purpose of the function 'MassShiftTransitions' is
# to generate a vector (FragmentMassShift) which contains
# the mass shift of a molecule(-fragment) of a given
# transition in relation to the same molecule(-fragment)
# containing no isotopes of higher mass. This is required to
# be able to yield the probability that a certain transition
# or full MS ion x is derived from another transition/full MS ion
# y due to natural stable isotope abundance. This probability is yielded
# by matching the mass shift k of transition x in FragmentMassShift
# with the mass shift k of transition y in CumProbList
# (this is the mass shift 'produced' by y through natural abundance).
# The probability is found associated with mass shift k in CumProbList.

MassShiftTransitions <- function(MoleculeInfo, MoleculeNo, Fragment) {
  
  Transitions <- MoleculeInfo[[MoleculeNo]][["Transitions"]]
  NumberTransitions <- nrow(Transitions)
  
  FragmentMassShift <- vector()
  
  for (TransitionNo in seq_len(NumberTransitions)) {
    FragmentMassShift[TransitionNo] <- Transitions[TransitionNo, Fragment]
  }  #TransitionNo
  
  return(FragmentMassShift)
  
}  #MassShiftTransitions

#' @importFrom magrittr '%>%'

RawDataExtraction <- function(data, MoleculeArray, logEnvironment, verbose) {
    
    if(verbose){message(date(), " :: processing raw data ...")}
    
    for (i in 2:ncol(data)) {
        data[, i] <- as.numeric(data[, i])
    }
    
    # remove rows that are ALL NA (=> don't remove individual NAs !)
    data <- data[!apply(apply(data, 1, is.na), 2, sum) == ncol(data), ]
    # remove columns that are all NA
    data <- data[, !apply(apply(data, 2, is.na), 2, sum) == nrow(data)]
    
    rownames(data) <- as.character(data[, "Measurements/Samples"])
    data <- as.matrix(data[, -which(colnames(data) %in% "Measurements/Samples"), drop = FALSE])
    txt_extracted <- rownames(data)
    
    # Extraction of measurement locations from the imported data to check for presence.
    
    MoleculesTotal <- length(MoleculeArray)
    MoleculeName <- names(MoleculeArray)
    
    MoleculeLocationList <- list()
    NumberTransitionsList <- list()
    TransitionLocationExpectedList <- list()
    TransitionLocationList <- list()
    MissingMoleculesList <- list()
    MissingTransitionsList <- list()
    
    for (MoleculeNo in seq_len(MoleculesTotal)) {
        
        MoleculeData <- MoleculeArray[[MoleculeNo]]  # entire 'MoleculeInfo' data for current molecule
        
        TransitionsExpected <- MoleculeData[["TransitionsExpected"]]  # from CalculateTransitions()
        NumberTransitions <- nrow(TransitionsExpected)
        
        # Check: Do the measurement labels in the first column of the data match the theoretically expected measurement labels?
        
        MissingTransitions.vec <- vector()
        TransitionLocationExpected.vec <- vector()
        
        for (TransitionNo in seq_len(NumberTransitions)) {
            
            tmpIdx <- stringr::str_detect(txt_extracted, paste0("^", rownames(TransitionsExpected)[TransitionNo], "$")) %>% which
            
            MissingTransitions.vec[TransitionNo] <- 0
            
            if (length(tmpIdx) == 0) {
                TransitionLocationExpected.vec[TransitionNo] <- 0
                MissingTransitions.vec[TransitionNo] <- 1
                
                notification <- stringr::str_c("In measurement data file: The expected measurement ID '", rownames(TransitionsExpected)[TransitionNo], "' could not be found in the measurement data file.\nCorrection will be performed, however the results may be less accurate. Please check your input file for typos.", 
                  "\nBe especially careful when considering fraction and mean enrichment values from molecules with missing measurements.")
                errorHandler(notification, logEnvironment, "warning", verbose=verbose)
                
            } else {
                
                TransitionLocationExpected.vec[TransitionNo] <- tmpIdx
                
            }
            
        }  #TransitionNo
        
        if (sum(TransitionLocationExpected.vec) == 0) {
            
            notification <- stringr::str_c("In measurement data file: None of the expected measurement IDs found for molecule '", names(MoleculeArray[MoleculeNo]), 
                "'.")
            errorHandler(notification, logEnvironment, "error", verbose=verbose)
            
        }
        
        names(TransitionLocationExpected.vec) <- rownames(TransitionsExpected)
        names(MissingTransitions.vec) <- rownames(TransitionsExpected)
        
        # If measurement IDs are missing, a new list of Measurement IDs is produced containing only those that were actually found in the input data.  This list
        # is then used for further computations.
        
        Transitions <- TransitionsExpected[MissingTransitions.vec == 0, , drop = FALSE]
        TransitionLocation <- TransitionLocationExpected.vec[MissingTransitions.vec == 0]
        
        TransitionLocationExpectedList[[MoleculeNo]] <- TransitionLocationExpected.vec
        TransitionLocationList[[MoleculeNo]] <- TransitionLocation
        
        MoleculeArray[[MoleculeNo]][["Transitions"]] <- Transitions
        MoleculeArray[[MoleculeNo]][["MissingTransitions"]] <- MissingTransitions.vec
        MoleculeArray[[MoleculeNo]][["TransitionLocationExpected"]] <- TransitionLocationExpected.vec
        MoleculeArray[[MoleculeNo]][["TransitionLocation"]] <- TransitionLocation
        
    }  #MoleculeNo
    
    if(verbose){message(date(), " :: processing raw data [OK]\n")}
    return(list(MoleculeInfo = MoleculeArray, dataRaw = data))
    
}  #RawDataExtraction

# Load input files, check their structure, extract information and check input logic

ExtractInputFileInformation <- function(ElementFile, MoleculeFile,
                                        MeasurementFile, UltraHighRes,
                                        CorrectTracerImpurity, logEnvironment,
                                        verbose) {
  ElementData <-
    checkElementDataStructure(
      data = readFileByExt(ElementFile, verbose = verbose), logEnvironment =
        logEnvironment, verbose = verbose
    )
  MoleculeData <-
    checkMoleculeDataStructure(
      data = readFileByExt(MoleculeFile, verbose = verbose),
      logEnvironment = logEnvironment, verbose = verbose
    )
  RawData <-
    checkRawData(
      inputFile = MeasurementFile, logEnvironment = logEnvironment,
      verbose = verbose
    )

  # Extract user input from the element information file

  ElementInfo <-
    ElementInfoExtraction(
      ElementData = ElementData,
      UltraHighRes = UltraHighRes,
      logEnvironment = logEnvironment,
      verbose = verbose
    )

  # Extract user input from the molecule information file

  MoleculeInfo <- MoleculeInfoExtraction(
    MoleculeData = MoleculeData,
    ElementInfo = ElementInfo,
    UltraHighRes = UltraHighRes,
    CorrectTracerImpurity = CorrectTracerImpurity,
    MoleculesFound = RawData$MoleculesFound,
    logEnvironment = logEnvironment,
    verbose = verbose
  )

  # Automatic determination of the transitions/measured isotopologues expected
  # according to the molecule parameters entered in the molecule information file

  if (!UltraHighRes) {
    MoleculeInfo <- CalculateTransitionsNR(
      MoleculeInfo = MoleculeInfo,
      ElementInfo = ElementInfo,
      verbose = verbose
    )
  } else {
    MoleculeInfo <- CalculateTransitionsUHR(
      MoleculeInfo = MoleculeInfo,
      ElementInfo = ElementInfo,
      verbose = verbose
    )
  }

  # Extraction of measurement information from the measurement file provided

  resultRawDataExtractionList <-
    RawDataExtraction(
      data = RawData[[1]],
      MoleculeArray = MoleculeInfo,
      logEnvironment = logEnvironment,
      verbose = verbose
    )

  MoleculeInfo <- resultRawDataExtractionList[["MoleculeInfo"]]

  MeasurementInfo <- resultRawDataExtractionList[["dataRaw"]]

  return(list(
    "MoleculeInfo" = MoleculeInfo,
    "ElementInfo" = ElementInfo,
    "MeasurementInfo" = MeasurementInfo
  ))
}

# Calculates the mean isotopic enrichment for the corrected values.


CalcMeanEnrichment <- function(MoleculeInfo, MoleculesTotal, SamplesTotal, roundDigits, CorrectionResultList, UltraHighRes, verbose) {
    
    if(verbose){message("\n", date(), " :: calculating mean enrichment ...")}
    
    MeanEnrichmentResultList <- list()
    
    for (SampleNo in seq_len(SamplesTotal)) {
        
        MeanEnrichmentSample <- list()
        
        for (MoleculeNo in seq_len(MoleculesTotal)) {
            
            MoleculeData <- MoleculeInfo[[MoleculeNo]]
            MoleculeName <- names(MoleculeInfo[MoleculeNo])
            CorrectionData <- CorrectionResultList[[SampleNo]][[MoleculeNo]]
            Transitions <- MoleculeData[["Transitions"]]
            NumberTransitions <- nrow(Transitions)
            
            if (UltraHighRes == FALSE) {
                
                MeanEnrichmentTmp <- 0
                
                for (TransitionNo in seq_len(NumberTransitions)) {
                  
                  if (is.na(CorrectionData[["CorrectedFractions"]][TransitionNo]) == FALSE) {
                    
                    # It is ok to use the mass shift values and not neccessarily the label values here, because dividing by TotalMassShiftMax will in the end lead to correct
                    # values
                    
                    TotalMassShift <- MoleculeData$Transitions$Precursor[[TransitionNo]]
                    
                    MeanEnrichmentTmp <- MeanEnrichmentTmp + as.numeric(CorrectionData[["CorrectedFractions"]][TransitionNo] * TotalMassShift)
                    
                  }
                  
                }  #TransitionNo
                
                TotalMassShiftMax <- max(MoleculeData$TransitionsExpected$Precursor)
                
                MeanEnrichmentSample[[MoleculeName]] <- round(MeanEnrichmentTmp/TotalMassShiftMax, digits = roundDigits)
                
            } else if (UltraHighRes) {
                
                Tracers <- MoleculeData$Fragment_1$Tracer
                
                for (Tracer in names(Tracers)) {
                  
                  MoleculeNamePlusTracer <- paste0(MoleculeName, "_", Tracer)
                  Label <- MoleculeData$Transitions[[Tracer]]
                  MeanEnrichmentTmp <- 0
                  
                  for (TransitionNo in seq_len(NumberTransitions)) {
                    
                    if (is.na(CorrectionData[["CorrectedFractions"]][TransitionNo]) == FALSE) {
                      
                      MeanEnrichmentTmp <- MeanEnrichmentTmp + CorrectionData[["CorrectedFractions"]][TransitionNo] * Label[[TransitionNo]]
                      
                    }
                    
                  }
                  
                  MeanEnrichmentSample[[MoleculeNamePlusTracer]] <- round(MeanEnrichmentTmp/Tracers[[Tracer]], digits = roundDigits)
                  
                }
                
            }
            
        }
        
        MeanEnrichmentResultList[[SampleNo]] <- MeanEnrichmentSample
        
    }
    
    names(MeanEnrichmentResultList) <- names(CorrectionResultList)
    
    MeanEnrichmentDataOutput <-
      as.data.frame(convertEnrichmentList2matrix(input = MeanEnrichmentResultList, verbose=verbose))
    
    if(verbose){message(date(), " :: calculating mean enrichment [OK]\n")}
    return(MeanEnrichmentDataOutput)
    
}  #MeanEnrichment

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{IsoCorrectoR}
\alias{IsoCorrectoR}
\title{IsoCorrectoR example data}
\format{A list with four elements, namely
\describe{
\item{tool_features: }{overview of correction features as provided by various tools IsoCorrectoR is compared with.}
\item{element_file: }{example data for an Element file. This information is independent of the resolution type (normal or high)}
\item{normal_resolution: }{list with two elements, containing examples for normal resolution data: molecule information and measurement data}
\item{high_resolution: }{list with two elements, containing examples for high resolution data: molecule information and measurement data}
}}
\usage{
data(IsoCorrectoR)
}
\description{
Example data for the IsoCorrectoR package
}
\author{
Paul Heinrich, Christian Kohler
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IsoCorrection.R
\name{IsoCorrection}
\alias{IsoCorrection}
\title{Algorithm For Natural Isotope Abundance And Tracer Purity Correction Of Data From Stable Isotope Labeling Experiments}
\usage{
IsoCorrection(MeasurementFile = NA, ElementFile = NA, MoleculeFile = NA,
  CorrectTracerImpurity = FALSE, CorrectTracerElementCore = TRUE,
  CalculateMeanEnrichment = TRUE, UltraHighRes = FALSE, DirOut = ".",
  FileOut = "result", FileOutFormat = "csv", ReturnResultsObject = TRUE,
  CorrectAlsoMonoisotopic = FALSE, CalculationThreshold = 10^-8,
  CalculationThreshold_UHR = 8, verbose = FALSE, Testmode = FALSE)
}
\arguments{
\item{MeasurementFile}{Required. The file that contains the measured data to
be corrected. Only ".xls",".xlsx" and ".csv" file formats are supported.}

\item{ElementFile}{Required. The file that contains the element information
required for correction. Only ".xls",".xlsx" and ".csv" file formats are
supported.}

\item{MoleculeFile}{Required. The file that contains the information on the
molecules for which data is to be corrected. Only ".xls",".xlsx" and ".csv"
file formats are supported.}

\item{CorrectTracerImpurity}{Logical. If TRUE, correction for isotopic
impurity of the tracer substrate is performed.}

\item{CorrectTracerElementCore}{Logical. If TRUE (recommended!), the tracer
element atoms in the core module (usually the part of the molecule that
does not come from derivatization) are considered when correcting.}

\item{CalculateMeanEnrichment}{Logical. If TRUE, the mean isotopic enrichment
is calculated for each molecule.}

\item{UltraHighRes}{Logical. If TRUE, high resolution correction is performed
on the data. Should only be set to TRUE, if you know that you have high
resolution data.}

\item{DirOut}{Character String. Defines the directory the corrected data and
log-file should be written to. Default directory is set to current working
directoy ('.').}

\item{FileOut}{Character String. Defines the name of the file that contains
the corrected data. The name of the fill will be
IsoCorrectoR_<FileOut>.<FileFormat>. If the format is set to "csv", the
name will also contain the type of the corrected data in the respective
file.}

\item{FileOutFormat}{Character String. Defines the format of the files that
contain the corrected data. Can either be "csv" or "xls". If set to "csv",
multiple files will be generated, one for each type of corrected data (eg.
corrected data, fractions, mean enrichment...). If set to "xls", all
correction results are provided in one excel file in different sheets.}

\item{ReturnResultsObject}{Logical. If TRUE, the correction results are
returned as a list in the current R_session in addition to writing the
results to a file. This is useful if the corrected data has to be further
processed directly in R.}

\item{CorrectAlsoMonoisotopic}{Logical. If TRUE, monoisotopic correction
results are also provided.}

\item{CalculationThreshold}{(Advanced Option) Numeric. Defines a threshold to
stop probability calculations at for making correction faster (normal
resolution mode). Should be left at the default value.}

\item{CalculationThreshold_UHR}{(Advanced Option) Numeric. Defines a
threshold to stop probability calculations at for making correction faster
(high resolution mode). Should be left at the default value.}

\item{verbose}{Logical. If TRUE, status messages are sent to standard output.}

\item{Testmode}{Logical. If TRUE, starts a testmode for development purposes.
Not required for users of IsoCorrectoR.}
}
\value{
The function returns a list with 4 elements

\describe{
 \item{success: }{string that is "TRUE" if the correction was successful, "FALSE" if an error occured and "WARNINGS"
 if warnings occured}
 \item{results: }{a list containing a dataframe for each type of corrected data (normal, fractions, mean enrichment ...).
 Will be NA if ReturnResultsObject is set to FALSE}
 \item{log: }{list containing log information on the correction run (parameters, file names and paths, warnings and errors)}
 \item{error: }{contains a string with the associated error message if an error occurred, empty otherwise}
}
}
\description{
\code{IsoCorrection} is the main function of the IsoCorrectoR package.
It performs the correction of mass spectrometry data from stable isotope
labeling experiments with regard to natural abundance and tracer purity.
Data from both MS and MS/MS experiments can be corrected
(with any tracer isotope: 13C, 15N, 18O...), as well as high resolution data
from multiple-tracer experiments (e.g. 13C and 15N used simultaneously).
}
\examples{

# Normal resolution data

   # 1) get path of example files
   path.molecule <- system.file("extdata","normal_resolution","MoleculeFile.csv",
      package = "IsoCorrectoR", mustWork = TRUE);
   path.element <- system.file("extdata","normal_resolution","ElementFile.csv",
      package = "IsoCorrectoR", mustWork = TRUE);
   path.measurement <- system.file("extdata","normal_resolution","MeasurementFile.csv",
      package = "IsoCorrectoR", mustWork = TRUE);

   # 2) run correction algorithm and save results in variable
   correctionResults <- IsoCorrection(MeasurementFile=path.measurement,
      ElementFile=path.element,
     MoleculeFile=path.molecule)

# High resolution data

   # 1) get path of example files
   path.molecule <- system.file("extdata","high_resolution","MoleculeFile.csv",
      package = "IsoCorrectoR", mustWork = TRUE);
   path.element <- system.file("extdata","high_resolution","ElementFile.csv",
      package = "IsoCorrectoR", mustWork = TRUE);
   path.measurement <- system.file("extdata","high_resolution","MeasurementFile.csv",
      package = "IsoCorrectoR", mustWork = TRUE);

   # 2) run correction algorithm and save results in variable
   correctionResults <- IsoCorrection(MeasurementFile=path.measurement,
      ElementFile=path.element,
     MoleculeFile=path.molecule,UltraHighRes=TRUE)

}
\references{
See Reference 1 \url{Link to IsoCorrectoR-Paper}
}
